/*-------------------------------------------------------------------------
|   rxtx is a native interface to serial ports in java.
|   Copyright 1997, 1998, 1999 by Trent Jarvi jarvi@ezlink.com.
|   Work originally contributed by david atkinson.
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
package gnu.io;

import java.io.IOException;
import java.io.BufferedInputStream;
import javax.comm.*;

class RXMonitor extends Thread {
	SerialInputStream Rt;
	RXTXPort	port;
	boolean 	CTS=false;
	boolean 	DSR=false;
	boolean 	RI=false;
	boolean 	CD=false;
	boolean		OUTPUT_BUFFER_EMPTY=false;
	boolean         switcher=false;
	boolean         debug=false;
	byte            OEtest=1;  //7
	byte            PEtest=2;  //8
	byte            BItest=4;  //9
	byte            FEtest=8;  //10
  		 
	public RXMonitor(SerialInputStream rt, RXTXPort port) {
		Rt=rt;
		this.port = port;
	}

	public void run() {
		byte flags;
		while (!port.thread_die) {
			flags=Rt.catchsigio();
			if(debug)
				showflags(flags);
			if((flags&FEtest)==FEtest) 
				RaiseEvent(SerialPortEvent.FE);
			if((flags&BItest)==BItest) 
				RaiseEvent(SerialPortEvent.BI);
			if((flags&PEtest)==PEtest) 
				RaiseEvent(SerialPortEvent.PE);
			if((flags&OEtest)==OEtest) 
				RaiseEvent(SerialPortEvent.OE);
			if(!isAvailable()){ 
				// monitor control lines and flowcontrol
				isCTS();
				isDSR();
				isRI();
				isCD();
			}
			// use sigio or just sleep for 1 second.
		}
		try{
		Thread.sleep(1);
		} catch(InterruptedException e){}
	}
	void showflags(int flags){
                        System.out.print("Flags = " + flags +" FEtest = " + FEtest +" & = " + (flags&FEtest));
                        System.out.print(" BItest = " + BItest +" & = " + (flags&BItest));
                        System.out.print(" PEtest = " + PEtest +" & = " + (flags&PEtest));
                        System.out.println(" OEtest = " + OEtest +" & = " + (flags&OEtest));
	}
	private boolean isAvailable() {
		SerialPortEvent SE;
		int DATA=SerialPortEvent.DATA_AVAILABLE;

		if  (((port.monitor & DATA)==0))
			return false;
		SE=new SerialPortEvent(port, DATA, false, true);
 		try {  
			if (Rt.nativeavailable(Rt.PortNumber) <= 0){
/* TRENT
*/
				SE=new SerialPortEvent(port, DATA, true, false);
				return(false);
			}
		} catch (IOException e) {}
		
		port.SPEventListener.serialEvent(SE);
		return(true);
	}
	private void isCD() {
		boolean RtisCD = Rt.isCD(Rt.PortNumber);
		SerialPortEvent SE;

		// check for change is CD state
		if ((CD == RtisCD)||(port.monitor & SerialPortEvent.CD)==0) 
			return;
		// System.out.println("CD state change ");
		// generate event if required
		SE=new SerialPortEvent(port, SerialPortEvent.CD, CD, RtisCD);
		port.SPEventListener.serialEvent(SE);
		// update hardware flow cntrl if required
		// if ((port.flowmode & SerialPort.FLOWCTRL_RTSCTS_IN)!=0)
		//      Rt.setRTSCTS(RtisCD);
		CD = RtisCD;
	}
	private void isRI() {
		boolean RtisRI=Rt.isRI(Rt.PortNumber);
		SerialPortEvent SE;

		if ((port.monitor & SerialPortEvent.RI)==0||(RI == RtisRI)) 
			return;
		SE=new SerialPortEvent(port, SerialPortEvent.RI, RI, RtisRI);
		port.SPEventListener.serialEvent(SE);
		RI = RtisRI;
	}
	private void isDSR() {
		boolean RtisDSR=Rt.isDSR(Rt.PortNumber);
		SerialPortEvent SE;

		if((port.monitor & SerialPortEvent.DSR)==0 ||(DSR == RtisDSR))
			return;
		SE=new SerialPortEvent(port, SerialPortEvent.DSR, DSR, RtisDSR);
		port.SPEventListener.serialEvent(SE);
		DSR = RtisDSR;
	}
	private void isCTS() {
		boolean RtisCTS=Rt.isCTS(Rt.PortNumber);
		SerialPortEvent SE;

		if ((port.monitor & SerialPortEvent.CTS)==0||(CTS == RtisCTS))
			return;
		SE=new SerialPortEvent(port, SerialPortEvent.CTS, CTS, RtisCTS);
		port.SPEventListener.serialEvent(SE);
		CTS = RtisCTS;
	}
	void RaiseEvent(int Event){
		SerialPortEvent SET;
		if(debug)
			System.out.println("Event " + Event); 
		SET=new SerialPortEvent(port, Event, true, true);
		port.SPEventListener.serialEvent(SET);
	}
	private void isEmpty() { /* interesting problem in the kernel */
		boolean RtisEmpty=Rt.isEmpty(Rt.PortNumber);
		SerialPortEvent SE;
		if(switcher)
			switcher=false;
		else
			switcher=true;

		if ((port.monitor & SerialPortEvent.OUTPUT_BUFFER_EMPTY)==0||(OUTPUT_BUFFER_EMPTY == switcher)) 
			return;
		SE=new SerialPortEvent(port, SerialPortEvent.OUTPUT_BUFFER_EMPTY, OUTPUT_BUFFER_EMPTY, switcher);
		port.SPEventListener.serialEvent(SE);
		OUTPUT_BUFFER_EMPTY = switcher;
	}
}
