/*-------------------------------------------------------------------------
|   A wrapper to convert RXTX into Linux Java Comm
|   Copyright 1998 Kevin Hester, kevinh@acm.org
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
package gnu.io;

import javax.comm.*;

import java.io.*;

/**
   This is the JavaComm for Linux driver.  
*/
public class RXTXCommDriver implements CommDriver {

	private final static boolean debug = false;

    /*
     * initialize() will be called by the CommPortIdentifier's static
     * initializer. The responsibility of this method is:
     * 1) Ensure that that the hardware is present.
     * 2) Load any required native libraries.
     * 3) Register the port names with the CommPortIdentifier.
	 * 
	 * <p>From the NullDriver.java CommAPI sample.
     */
    public void initialize()
        {
		for(int i = 0; i < portNames.length; i++) {
			CommPortIdentifier.addPortName(portNames[i],
				CommPortIdentifier.PORT_SERIAL, this);
        }
		}
        
    /*
     * getCommPort() will be called by CommPortIdentifier from its openPort()
     * method. portName is a string that was registered earlier using the
     * CommPortIdentifier.addPortName() method. getCommPort() returns an
     * object that extends either SerialPort or ParallelPort.
	 *
	 * <p>From the NullDriver.java CommAPI sample.
     */
    public CommPort getCommPort(String portName, int portType)    
        {
		if( debug ) {
			System.out.println("RXTXCommDriver.getCommPort("+portName+")");
		}

		//map dos port names to something reasonable really need to
                //check which OS so we can map the dos stuff to non linux stuff
                //too
		for (int i=0 ; i< dosPortNames.length ; i++) {
			if (portName.equals(dosPortNames[i]))
				portName = portNames[i+1];
		}
		try {
			if( debug ) {
				System.out.println("opening:"+portName);
			}
			return new RXTXPort(portName);
		}
		catch(IOException e)
		{
			if( debug ) 
				System.out.println("Can't throw exception - ignoring: " + e);
			return null;
		}
        }

	/**
	   A query function to provide runtime version checking
	*/
	public static String getVersion()
	    {
		return "JCL Version 1.1 - Copyright 1998 Kevin Hester, kevinh@acm.org - GPL dist ok";
		}

	private final static String dosPortNames[] =
		{"COM1", "COM2", "COM3", "COM4"};


	private final static String portNames[] =
	    {
		"/dev/modem",
		"/dev/ttyS0",
		"/dev/ttyS1",
		"/dev/ttyS2",
		"/dev/ttyS3",

		"/dev/ttyd0",   // irix Serial Ports
		"/dev/ttyd1",
		"/dev/ttyd2",
		"/dev/ttyd3",

		"/dev/ttyf0",   // irix Serial Ports with hardware flow
		"/dev/ttyf1",
		"/dev/ttyf2",
		"/dev/ttyf3",

		"/dev/ttym0",  // irix Modems
		"/dev/ttym1",
		"/dev/ttym2",
		"/dev/ttym3",

		"/dev/ttyq0",  // irix Pseudo tty's
		"/dev/ttyq1",
		"/dev/ttyq2",
		"/dev/ttyq3",

		"/dev/ttyW0",  // specialix cards on linux
		"/dev/ttyW1",
		"/dev/ttyW2",
		"/dev/ttyW3",
		"/dev/ttyW4",
		"/dev/ttyW5",
		"/dev/ttyW6",
		"/dev/ttyW7",
		"/dev/ttyW8",
		"/dev/ttyW9",
		"/dev/ttyW10",
		"/dev/ttyW11",  // if you add more than 12 here you better
				// edit SerialImp.c and increase MAXPORTS

		dosPortNames[0], 
		dosPortNames[1], 
		dosPortNames[2], 
		dosPortNames[3]    	
		};

	}
