/*-------------------------------------------------------------------------
|   rxtx is a native interface to serial ports in java.
|   Copyright 1997, 1998, 1999 by Trent Jarvi jarvi@ezlink.com.
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------
   win32 support has been removed.  There was not enough developer interest
   if your interested in working on win32 support contact jarvi@ezlink.com
--------------------------------------------------------------------------*/
#include "config.h"
#include "gnu_io_SerialOutputStream.h"
#include "gnu_io_SerialInputStream.h"
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/time.h>
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_SYS_FCNTL_H
#   include <sys/fcntl.h>
#endif
#ifdef HAVE_SYS_FILE_H
#   include <sys/file.h>
#endif
#ifdef HAVE_SYS_SIGNAL_H
#   include <sys/signal.h>
#endif
#ifdef EVENTNOTIFICATION
#if defined(__linux__) /* this will change */
#   include <linux/serial.h>
#endif
#endif

extern int errno;
static int Initialized=0;  /* used to prevent reinitialization of 
                              the commport (cp) array.  */
#ifdef ASYNC
struct sigaction saio;
#endif

#ifdef TIMEOUT
#define WAYNE  /* Waynes Roberts work towards better timeouts */
#endif
#undef GARBAGE
#define MAX_SKIP_BUFFER_SIZE (64*1024)
#define MAXPORTS 12  /* input/output share the same struct */

#include "SerialImp.h"

int sIGiO;


void signal_handler (int status);

/*----------------------------------------------------------
SerialOutputStream_catchsigio

   accept:      none
   perform:     sleep till a signal is recieved.  The signal handler 
                gives us which port in sIGiO.  So we check that port
                for goodies and return them.
   return:      in with events that happened.
   exceptions:  none
   comments:    this is probably not going to work with multiple ports
                without some sort of cross talk...  Hmm.  How does the
                rxmonitor know the events are ment for its port?.
----------------------------------------------------------*/ 
JNIEXPORT jbyte JNICALL Java_gnu_io_SerialInputStream_catchsigio (
	JNIEnv *Env, 
	jobject jobj
	)
{
	jbyte ret=0;
	struct commport  *Port=0;
#ifdef TIMEOUTS
#ifdef DEBUG
	usleep(1000000);   /* sleep for 1 second */
#else /* DEBUG */
	usleep(1);  /* sleep for 1 microsecond */
#endif
#else
	sleep(1);  /* sleep for 1 second */
#endif /* TIMEOUTS */
	if(sIGiO>=0){ 
		Port=&cp[sIGiO];
#ifdef ASYNC
#ifdef EVENTNOTIFICATION
#if defined(__linux__) /* this will change */
#ifdef DEBUG
		printf("sIGiO=%s\n",Port->name);
#endif
#ifdef FULL_EVENT
		if(Port->sis.frame!=Port->osis.frame)
			ret|=FRAME_ERROR;
		if(Port->sis.brk!=Port->osis.brk)
			ret|=BREAK_ERROR
		if(Port->sis.parity!=Port->osis.parity)
			ret|=PARITY_ERROR
		if(Port->sis.overrun!=Port->osis.overrun)
			ret|=OVERRUN_ERROR
#endif
#endif
#endif
#endif
		sIGiO=-1;
	}
	return(ret);
}
/*----------------------------------------------------------
SerialOutputStream_InitPortArray

   accept:      none
   perform:     Set the structures in the commport array to 
                sane values.  Set initialized global value
                so we dont initialize after opening ports.
   return:      none.
   exceptions:  none
   comments:
----------------------------------------------------------*/ 
JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_InitPortArray (
	JNIEnv *Env, 
	jobject jobj
	)
{
	int i;
	if(!Initialized) {
		Initialized=1;
		for(i=0;i<MAXPORTS;i++){
			cp[i].available=1;
			cp[i].fd_is_needed=JNI_TRUE;
			cp[i].name[0]='\0';
			cp[i].fd=-1;
			cp[i].number_of_times_opened=0;
			cp[i].no_block=BLOCKING;
			cp[i].hardwarefc=JNI_FALSE;
#ifdef WAYNE
			cp[i].JTIMEOUT= 0;
			cp[i].THRESHOLD=0;
			cp[i].BAUDRATE=9600;
#endif
			cp[i].last_TIOCMGET_result=0;
			cp[i].OutputBufferSize=1;
			cp[i].InputBufferSize=1;
			cp[i].FrameByte=0;
		}
#ifdef ASYNC
	saio.sa_handler = signal_handler; 
	sigemptyset(&saio.sa_mask);
	saio.sa_flags = 0;
#if defined (__linux__)
	saio.sa_restorer = NULL;
#endif
	sigaction(SIGIO,&saio,NULL);
#ifdef REPORTEMPTY
	sigaction(SIGCONT,&saio,NULL);
#endif
#endif
	Initialized=2;
	}
}
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_InitPortArray ( JNIEnv *Env, jobject jobj)
{Java_gnu_io_SerialOutputStream_InitPortArray(Env, jobj);}

/*----------------------------------------------------------
SerialOutputStream_open

   accept:      The device to open.  ie "/dev/ttyS0"
   perform:     open the device, set the termios struct to sane settings and 
                set the filedescriptor (jfd) on the java side.  Enable ASYNC
                if required.
   return:      PortNumber.
   exceptions:  open, tcgetattr and tcsetattr fails 
   comments:    Very often people complain about not being able to get past
                this function and it turns out to be permissions on the 
                device file or bios has the device disabled.

----------------------------------------------------------*/ 
#define OPENERROR "OutputStream_open\nThis is a common problem.  Either the device does not physically exist or you do not have permission to access it."


JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_open (
	JNIEnv *Env, 
	jobject jobj, 
	jstring jstr
   ) 
{
	int       FileDescriptor  = -1;
	jclass    jcls = GetObjectClass(Env, jobj);
	jfieldID  jfd  = GetStaticFieldID(Env, jcls, "FileD", "I");
	int PortNumber=0;
	int Excess=JNI_FALSE;
	struct commport *Port=0;

	const char *filename=GetStringUTFChars(Env,jstr,0);
	ReleaseStringUTFChars(Env, jstr, NULL);
	
#ifdef TIMEOUTS
	while(Initialized<2) usleep(1);
#else
	while(Initialized<2) sleep(1);
#endif
	Initialized=1;
	while((cp[PortNumber].available < 0) && 
		(PortNumber<MAXPORTS) && 
		(strcmp(cp[PortNumber].name,filename)!=0))
	{
		PortNumber++;
		printf("Portnumber++ %i\n",PortNumber);
	}
	Port=&cp[PortNumber];
	if(Port->number_of_times_opened>1){
		errno=EBADFD;
		Excess=JNI_TRUE;
		goto fail;
	}
	if(strcmp(Port->name,filename)==0) goto succed;

	toss_debug("Open", PortNumber);

	FileDescriptor = open(filename, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if(FileDescriptor < 0) { errno=EBADFD;goto fail; }
	Port->fd=FileDescriptor;
	sprintf(Port->name,"%s",filename); 
	SetStaticIntField(Env, jcls, jfd, Port->fd);

	if(tcgetattr(Port->fd, &Port->ttyorig)< 0) goto fail;
	makeraw(&Port->ttyset);
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	if(cfsetispeed(&Port->ttyset, B19200)< 0) goto fail;
	if(cfsetospeed(&Port->ttyset, B19200)< 0) goto fail;
#endif /* newer libc in linux */
#ifdef __FreeBSD__
	if(cfsetspeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail;

	toss_debug("Open Results", PortNumber);
#ifdef ASYNC
#ifdef EVENTNOTIFICATION
#if defined(__linux__) /* this will change */
        ioctl(Port->fd, TIOCGICOUNT,&Port->osis);
#endif
#endif /* EVENTNOTIFICATION */
	fcntl(Port->fd, F_SETOWN, getpid());
	fcntl(Port->fd, F_SETFL, FASYNC);
#endif /* ASYNC */

succed:
	Port->available=-1; 
	SetStaticIntField(Env, jcls, jfd, Port->fd);
	Port->number_of_times_opened++;
	Initialized=2;
	Port->fd_is_needed=JNI_FALSE;
	return(PortNumber);
fail:
	if(!Excess){
		Port->fd=-1;
		Port->name[0]='\0';
	}
	Initialized=2;
	toss_error(Env,"open",strerror(errno));
	return(-1);

}
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_open ( JNIEnv *Env, jobject jobj, jstring jstr) 
   { return(Java_gnu_io_SerialOutputStream_open(Env, jobj, jstr)); }

/*----------------------------------------------------------
SerialOutputStream_openAppend

   accept:      string of device to open in append mode.  
   perform:     IOexception.
   return:      none.
   exceptions:  default.
   comments:    just call open.  who knows...   
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_openAppend (
	JNIEnv *Env, 
	jobject jobj, 
	jstring jstr
   ) 
{
	Java_gnu_io_SerialOutputStream_open (Env, jobj, jstr);
}
/*----------------------------------------------------------
SerialOutputStream_nativewrite

   accept:      ji byte to write. (passed as int)
   perform:     write a java int to the device.
   return:      none.
   exceptions:  fd is invalid
                write failed.
   comments:    
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_nativewrite (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jint ji
   ) 
{
	struct commport *Port=&cp[PortNumber];
	unsigned char byte=ji;

	toss_debug("write", PortNumber);
	if (Port->fd > 0) {
		if( write(Port->fd, &byte, sizeof(unsigned char)) <0) goto fail;
	}
	else  {errno= EBADFD; goto fail;}
	return;
fail:
	IOException(Env,"write", strerror(errno));
	
}
/*----------------------------------------------------------
SerialOutputStream_writeBytes

   accept:      jbarray: bytes used for writing 
                Offset:  Number of bytes to skip too.  (confusing with serial)
                Length:  Number of bytes to write
   perform:     write zeros up to offset.  write length bytes of jbarray
   return:      none.
   exceptions:  No file descriptor.
                Write fail.
   comments:
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_writeBytes (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jbyteArray jbarray, 
	jint Offset, 
	jint Length 
   )
{
	jbyte    *body = GetByteArrayElements(Env, jbarray, 0);
	int i;
	struct commport *Port=&cp[PortNumber];
	unsigned char *bytes = ( unsigned char *) malloc (Length);
	for(i=0;i<Length;i++) bytes[i]=body[i];
	ReleaseByteArrayElements(Env, jbarray, body, 0);

	printf("writeBytes Length = %i Offset %i\n",Length, Offset);

	toss_debug("writeBytes", PortNumber);
	if (Port->fd < 0)  {errno= EBADFD ; goto fail; }
	for(i=0;i<Offset;i++) {
		if( write(Port->fd, "0", 1) <0) goto fail;
	}
	if( write(Port->fd, bytes, sizeof(unsigned char)*Length) <0) goto fail;
	free(bytes);
	return;
fail:
	free(bytes);
	ReleaseByteArrayElements(Env, jbarray, body, 0);
	IOException(Env,"writeBytes", strerror(errno));
}
/*----------------------------------------------------------
SerialOutputStream_close

   accept:      none.
   perform:     get the fd from the java end and close it.
   return:      none.
   exceptions:  close fails
   comments:
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_close (
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber
   ) 
{
	struct commport *Port=&cp[PortNumber];

#ifdef TIMEOUTS
	while(Initialized<2 || Port->fd_is_needed) usleep(1);
#else
	while(Initialized<2 || Port->fd_is_needed) sleep(1);
#endif
	Initialized=1;
	Port->fd_is_needed=JNI_TRUE;

	toss_debug("close", PortNumber);
	if(Port->available==-1){
		if(--Port->number_of_times_opened ==0){
			if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyorig) <0) {
				toss_error(Env,"close",strerror(errno));
			}
			if(close(Port->fd)<0) goto fail;
			Port->available=1; 
			Port->fd=-1;
			Port->name[0]='\0';
		}
	}
	Port->fd_is_needed=JNI_FALSE;
	Initialized=2;
	return;
fail:
	Initialized=2;
	IOException(Env,"close", strerror(errno));
	return;
}
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_close ( JNIEnv *Env, jobject jobj, jint PortNumber)
   { Java_gnu_io_SerialOutputStream_close(Env, jobj, PortNumber); }
/*----------------------------------------------------------
SerialOutputStream_tcsetattr

   accept:      actions.
   perform:     Get the termios structure to the kernel.
   return:      1 on success.
                0 on failure.
   exceptions:  Bad File Descriptor.
   comments:    This is when the changes actually take place on the device.
                FIXME actions should be passed to tcsetattr?  check this.
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcsetattr ( 
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
        jint actions
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcsetattr", PortNumber);
	if (Port->fd < 0) goto fail;
	return(tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset));

fail:
	IOException(Env, "tcsetattr", strerror(EBADFD));
	return(-1);
}
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_tcsetattr ( JNIEnv *Env, jobject jobj, jint PortNumber, jint actions)
   { return(Java_gnu_io_SerialOutputStream_tcsetattr(Env,jobj,PortNumber,actions)); }
/*----------------------------------------------------------
SerialOutputStream_tcsendbreak

   accept:     duration.  FIXME? (actual time is 1/4 to 1/2 duration)
   perform:    send 0 bytes for actual time.  not less than 0.25 seconds.
               if the device is async then no action. (not controllable)
   return:     0 success
               -1 fail
   exceptions: Bad File Descriptor.
   comments:   Have to check... wasnt java sending duration as seconds?
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcsendbreak ( 
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
        jint duration 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcsendbreak", PortNumber);
	if (Port->fd < 0) goto fail;
	return( tcsendbreak ( (int) Port->fd, (int) duration ));

fail:
	IOException(Env, "tcsendbreak", strerror(EBADFD));
	return(-1);
}

/*----------------------------------------------------------
SerialOutputStream_tcdrain

   accept:      none.
   perform:     return after all pending data is written.
   return:      0 on success.
                -1 on error.
   exceptions:  Bad FileDescriptor.
   comments:
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcdrain ( 
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcdrain", PortNumber);

	if (Port->fd < 0) goto fail;
     	return( tcdrain ( (int) Port->fd ));

fail:
	IOException(Env, "tcdrain", strerror(EBADFD));
	return(-1);
}

/*----------------------------------------------------------
SerialOutputStream_tcflush

   accept:      selector.  Type of action.
   perform:     flush input/output depending on selector.     
                   TCIFLUSH flush input.
                   TCOFLUSH flush output
                   TCIOFLUSH flush input and output
   return:      0 on success.
                -1 on failure
   exceptions:  Bad FileDescriptor.
   comments:    output streams should not flush input?
                should we check?
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcflush ( 
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
        jint selector 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcflush", PortNumber);
	if (Port->fd < 0) goto fail;
	return( tcflush ( (int) Port->fd, (int) selector ));

fail:
	IOException(Env, "tcflush", strerror(EBADFD));
	return(-1);
}

/*----------------------------------------------------------
SerialOutputStream_tcflow

   accept:      action.
   perform:     try to start/stop input/output.
                   TCOOFF stop output.
                   TCOON  restart output.
                   TCIOFF ask remote device to stop sending
                   TCION  ask remote device to restart sending
   return:      0 on success.
                -1 on failure.
   exceptions:  bad FileDescriptor.
   comments:    output streams probably should not control input.
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcflow ( 
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
        jint action 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcflow", PortNumber);

	if (Port->fd < 0) goto fail;
	return( tcflow ( (int) Port->fd, (int) action ));

fail:
	IOException(Env, "tcflow", strerror(EBADFD));
	return(-1);
}
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_tcflow ( JNIEnv *Env, jobject jobj, jint PortNumber, jint action )
   { return( Java_gnu_io_SerialInputStream_tcflow(Env,jobj, action, PortNumber)); }
/*----------------------------------------------------------
SerialOutputStream_cfmakeraw

   accept:      none.
   perform:     set the termios structure into a sane mode.
                   c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|
                                ICRNL|IXON);
                   c_oflag &= ~OPOST;
                   c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                   c_cflag &= ~(CSIZE|PARENB);
                   c_cflag |= CREAD;  CLOCAL can be ored to ignore modem control lines 
                   c_cflag |= CS8;
                   c_cc[VSUSP] = -1;

   return:      0 on success.
                -1 on failure.
   exceptions:  none.
   comments:    makeraw()  appears to be a linux/gnu specific function.  We
                implement it hear to avoid problems/confusion.
                Check makeraw() below to make sure 'perform' is current.
                Obviously OutputStream is changing Input stream values...
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_cfmakeraw (
	JNIEnv *Env, 
	jobject jobj ,
	jint PortNumber
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("cfmakeraw", PortNumber);
	makeraw(&Port->ttyset);
}
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_cfmakeraw ( JNIEnv *Env, jobject jobj, jint PortNumber ) 
   { Java_gnu_io_SerialOutputStream_cfmakeraw(Env, jobj, PortNumber); } 
/*----------------------------------------------------------
SerialOutputStream_isDSR
   accept:      none.
   perform:     check if TIOCM_DSR is set
   return:      1 if TIOCM_DSR is set
                0 if TIOCM_DSR is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                DSR stands for Data Set Ready.
----------------------------------------------------------*/
JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isDSR (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	struct commport *Port=&cp[PortNumber];
        unsigned int result;

	if(Port->fd_is_needed||Port->fd<0) return(JNI_FALSE);
	Port->fd_is_needed=JNI_TRUE;
		ioctl (Port->fd, TIOCMGET, &result);
	Port->fd_is_needed=JNI_FALSE;
	return (result & TIOCM_DSR ? JNI_TRUE : JNI_FALSE);
}

/*----------------------------------------------------------
SerialInputStream_isFrame
   accept:      none.
   perform:     nothing.
   return:      True if a frame error has occured (in theory)
   exceptions:  Bad FileDescriptor.
   comments:    Not used.  may be used if ioctls are available/found.
                see signal_handler() for how this is done in linux
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isFrame (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	if (cp[PortNumber].fd < 0) goto fail;
#ifdef ASYNC
	return (JNI_TRUE);
#else
	return (JNI_FALSE);
#endif

fail:
	IOException(Env, "isFrame", strerror(EBADFD));
	return(JNI_FALSE);
}
/*----------------------------------------------------------
SerialInputStream_isEmpty
   accept:      none.
   perform:     nothing.
   return:      True if an output is done (in theory)
   exceptions:  Bad FileDescriptor.
   comments:    Not used.  may be used if ioctls are available/found.
                see signal_handler() for how this is done in linux
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isEmpty (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	if (cp[PortNumber].fd < 0) goto fail;
#ifdef ASYNC
	return (JNI_TRUE);
#else
	return (JNI_FALSE);
#endif
fail:
	IOException(Env, "isEmpty", strerror(EBADFD));
	return(JNI_FALSE);
}
/*----------------------------------------------------------
SerialInputStream_isOverrun
   accept:      none.
   perform:     nothing.
   return:      True if an overrun has occured (in theory)
   exceptions:  Bad FileDescriptor.
   comments:    Not used.  may be used if ioctls are available/found.
                Interesting problem.  Currently not done anywhere.
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isOverrun (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	if (cp[PortNumber].fd < 0) goto fail;
#ifdef ASYNC
	return (JNI_TRUE);
#else
	return (JNI_FALSE);
#endif

fail:
	IOException(Env, "isOverrun", strerror(EBADFD));
	return(JNI_FALSE);
}
/*----------------------------------------------------------
SerialInputStream_isBreak
   accept:      none.
   perform:     nothing.
   return:      True if a break has occured (in theory)
   exceptions:  Bad FileDescriptor.
   comments:    Not used.  may be used if ioctls are available/found.
                see signal_handler() for how this is done in linux
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isBreak (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	if (cp[PortNumber].fd < 0) goto fail;
#ifdef ASYNC
	return (JNI_TRUE);
#else
	return (JNI_FALSE);
#endif

fail:
	IOException(Env, "isBreak", strerror(EBADFD));
	return(JNI_FALSE);
}
/*----------------------------------------------------------
SerialInputStream_isParity
   accept:      none.
   perform:     nothing.
   return:      True if a parity error has occured (in theory)
   exceptions:  Bad FileDescriptor.
   comments:    Not used.  may be used if ioctls are available/found.
                see signal_handler() for how this is done in linux
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isParity (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
	if (cp[PortNumber].fd < 0) goto fail;
#ifdef ASYNC
	return (JNI_TRUE);
#else
	return (JNI_FALSE);
#endif
fail:
	IOException(Env, "isParity", strerror(EBADFD));
	return(JNI_FALSE);
}

/*----------------------------------------------------------
SerialInputStream_isCD
   accept:      none.
   perform:     check if TIOCM_CD is set
   return:      1 if TIOCM_CD is set
                0 if TIOCM_CD is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                CD stands for Carrier Detect.
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isCD (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
	) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if(Port->fd_is_needed||Port->fd<0) return(JNI_FALSE);
	Port->fd_is_needed=JNI_TRUE;
        	ioctl (Port->fd, TIOCMGET, &result);
	Port->fd_is_needed=JNI_FALSE;
	return (result & TIOCM_CD ? JNI_TRUE : JNI_FALSE);

}
/*----------------------------------------------------------
SerialInputStream_isCTS
   accept:      none.
   perform:     check if TIOCM_CTS is set
   return:      1 if TIOCM_CTS is set
                0 if TIOCM_CTS is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                CTS stands for Clear To Send.
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isCTS (
        JNIEnv *Env, 
	jobject jobj ,
	jint PortNumber
	) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if(Port->fd_is_needed||Port->fd<0) return(JNI_FALSE);
	Port->fd_is_needed=JNI_TRUE;
        	ioctl (Port->fd, TIOCMGET, &result);
	Port->fd_is_needed=JNI_FALSE;
	return (result & TIOCM_CTS ? JNI_TRUE : JNI_FALSE);

}

/*----------------------------------------------------------
SerialInputStream_isRI
   accept:      none.
   perform:     check if TIOCM_RI is set
   return:      1 if TIOCM_RI is set
                0 if TIOCM_RI is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                RI stands for Ring Indicator.
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isRI (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if(Port->fd_is_needed||Port->fd<0) return(JNI_FALSE);
	Port->fd_is_needed=JNI_TRUE;
        	ioctl (Port->fd, TIOCMGET, &result);
	Port->fd_is_needed=JNI_FALSE;
	return (result & TIOCM_RI ? JNI_TRUE : JNI_FALSE);
}

/*----------------------------------------------------------
SerialInputStream_isRTS
   accept:      none.
   perform:     check if TIOCM_RTS is set
   return:      1 if TIOCM_RTS is set
                0 if TIOCM_RTS is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                tcgetattr with c_cflag CRTS_IFLOW

----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isRTS (
        JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if(Port->fd_is_needed||Port->fd<0) return(JNI_FALSE);
	Port->fd_is_needed=JNI_TRUE;
        	ioctl (Port->fd, TIOCMGET, &result);
	Port->fd_is_needed=JNI_FALSE;
	return (result & TIOCM_RTS ? JNI_TRUE : JNI_FALSE);
}

/*----------------------------------------------------------
SerialOutputStream_setRTS
   accept:      state  flag to set/unset.
   perform:     depends on the state flag
                if 1 TIOCM_RTS is set
                if 0 TIOCM_RTS is unset
   return:      none.
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                tcsetattr with c_cflag CRTS_IFLOW
----------------------------------------------------------*/

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_setRTS (
        JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jboolean state 
) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if (Port->fd < 0) goto fail;
        ioctl (Port->fd, TIOCMGET, &result);
        if (state == JNI_TRUE)
                result |= TIOCM_RTS;
        else
                result &= ~TIOCM_RTS;
        ioctl (Port->fd, TIOCMSET, &result);
	return;
fail:
	IOException(Env, "setRTS", strerror(EBADFD));
}
/*----------------------------------------------------------
SerialInputStream_isDTR
   accept:      none.
   perform:     check if TIOCM_DTR is set
   return:      1 if TIOCM_DTR is set
                0 if TIOCM_DTR is not set
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                DTR stands for Data Terminal Ready.
----------------------------------------------------------*/

JNIEXPORT jboolean JNICALL Java_gnu_io_SerialOutputStream_isDTR (
        JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber 
) 
{
	unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if (Port->fd < 0) goto fail;
	ioctl (Port->fd, TIOCMGET, &result);
	return (result & TIOCM_DTR ? JNI_TRUE : JNI_FALSE);

fail:
	IOException(Env, "isDTR", strerror(EBADFD));
	return(-1);
}
/*----------------------------------------------------------

SerialInputStream_setDTR
   accept:      state  flag to set/unset.
   perform:     depends on the state flag
                if 1 TIOCM_DTR is set
                if 0 TIOCM_DTR is unset
   return:      none.
   exceptions:  Bad FileDescriptor.
   comments:    This appears to be legitimate use of ioctl.
                DTR stands for Data Terminal Ready.
----------------------------------------------------------*/

JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_setDTR (
        JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jboolean state 
) 
{
        unsigned int result;
	struct commport *Port=&cp[PortNumber];

	if (Port->fd < 0) goto fail;
        ioctl (Port->fd, TIOCMGET, &result);
        if (state == JNI_TRUE)
                result |= TIOCM_DTR;
        else
                result &= ~TIOCM_DTR;
        ioctl (Port->fd, TIOCMSET, &result);
	return;
fail:
	IOException(Env, "setDTR", strerror(EBADFD));
}

/*----------------------------------------------------------
SerialOutputStream_tcgetpgrp

   accept:      none.
   perform:     gets process group ID.
   return:      process group ID on success
                -1 on error
   exceptions:  Bad FileDescriptor.
   comments:    
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcgetpgrp ( 
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
   )
{
	struct commport *Port=&cp[PortNumber];
	toss_debug("tcgetpgrp", PortNumber);
	if (Port->fd < 0) goto fail;
	return(tcgetpgrp ( (int) Port->fd ));

fail:
	IOException(Env, "tcgetpgrp", strerror(EBADFD));
	return(-1);
}
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_tcgetpgrp ( JNIEnv *Env, jobject jobj, jint PortNumber )
   { return(Java_gnu_io_SerialOutputStream_tcgetpgrp(Env, jobj,PortNumber)); }
/*----------------------------------------------------------
SerialOutputStream_tcsetpgrp

   accept:      Process group ID.
   perform:     sets process group id.
   return:      Process group ID on success.
                -1 on error.
   exceptions:  Bad FileDescriptor.
   comments:    
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_tcsetpgrp ( 
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jint pgrpid 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("tcsetpgrp", PortNumber);
	if (Port->fd < 0) goto fail;
	return( tcsetpgrp ( (int) Port->fd, (int) pgrpid ));

fail:
	IOException(Env, "tcsetpgrp", strerror(EBADFD));
	return(-1);


}
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_tcsetpgrp ( JNIEnv *Env, jobject jobj,jint PortNumber, jint pgrpid )
   { return(Java_gnu_io_SerialOutputStream_tcsetpgrp(Env, jobj,PortNumber,pgrpid)); }

/*----------------------------------------------------------
SerialOutputStream_cfgetospeed

   accept:      none.
   perform:     Checks the termios structure for the current output speed.
   return:      Output speed as defined in termbits.h.
   exceptions:  none.
   comments:    the termios structure does not represent what the kernel has
                unless tcsetattr/tcgetattr has been called.  ie the structure
                is just a holder of info.  you use it to get info from the
                kernel and set values in the kernel

                low level speed control
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_cfgetospeed (
	JNIEnv *Env, 
	jobject jobj ,
	jint PortNumber
   )
{
	struct commport *Port=&cp[PortNumber];

	if (Port->fd < 0) goto fail;
	toss_debug("cfgetospeed", PortNumber);
	return( cfgetispeed ( &Port->ttyset ));

fail:
	IOException(Env, "tcsetpgrp", strerror(EBADFD));
	return(-1);
}
/*----------------------------------------------------------
SerialOutputStream_cfsetospeed

   accept:      termbits.h defined value for speed.
   perform:     set the termios structure output speed value.
   return:      0 on success.
                -1 on failure.
   exceptions:  none.
   comments:    the termios structure is not what the kernel is using.  use
                tcsetattr to get the kernel to read the speed after setting
                it here.
                B0 is 'hang up'

                low level speed control
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_cfsetospeed (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jint speed
   )
{
	struct commport *Port=&cp[PortNumber];

	if (Port->fd < 0) goto fail;
	toss_debug("cfsetospeed", PortNumber);
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	return( cfsetospeed ( &Port->ttyset, (int) speed));
#endif
#ifdef __FreeBSD__
	return( cfsetspeed ( &Port->ttyset, (int) speed));
#endif
fail:
	IOException(Env, "tcsetpgrp", strerror(EBADFD));
	return(-1);
}
/*----------------------------------------------------------
SerialOutputStream_GetSpeed

   accept:     none.
   perform:    get the speed from the termios structure
   return:     the speed in bits-per-second.
   exceptions: none.
   comments:   high level speed control.  called from
                  RXTXPort.java
                  SerialOutputStream.java
                  Test.java
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_GetSpeed (
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("GetSpeed", PortNumber);
	if (Port->fd < 0) goto fail;
	return(speed_t2bps(cfgetospeed(&Port->ttyset)));
fail:
	IOException(Env, "tcsetpgrp", strerror(EBADFD));
	return(-1);
}
/*----------------------------------------------------------
SerialOutputStream_SetSpeed

   accept:     speed in bits-per-second.  See bps2speed_t for 
               valid speeds.  
   perform:    set the termios structure speed to speed.
   return:     none.
   exceptions: If fd is <1 bad file descriptor
               tcgetattr tcgetspeed, tcsetspeed and tcsetattr failures. 
   comments:   high level speed control.  called from
                  RXTXPort.java
                  SerialOutputStream.java
                  Test.java
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_SetSpeed (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
        jint speed
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("SetSpeed", PortNumber);
	if (Port->fd < 0) { errno=EBADFD;goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail; 
	if (bps2speed_t(speed)<=0){ errno=EINVAL;goto fail; }
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	if(cfsetospeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif
#ifdef __FreeBSD__
	if(cfsetspeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif
	if(tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail; 
#ifdef WAYNE
	Port->BAUDRATE = speed;
#endif
	return;
fail:
	IOException(Env,"SetSpeed", strerror(errno));
}
/*----------------------------------------------------------
SerialInputStream_cfgetispeed

   accept:      none.
   perform:     Checks the termios structure for the current input speed.
   return:      Input speed as defined in termbits.h.
   exceptions:  none.
   comments:    the termios structure does not represent what the kernel has
                unless tcsetattr/tcgetattr has been called.  ie the structure
                is just a holder of info.  you use it to get info from the
                kernel and set values in the kernel

                low level speed control
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_cfgetispeed (
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber 
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("cfgetispeed", PortNumber);
	if (Port->fd < 0) goto fail;
	return( cfgetispeed ( &Port->ttyset ));
fail:
	IOException(Env,"SetSpeed", strerror(errno));
	return(-1);
}
/*----------------------------------------------------------
SerialInputStream_cfsetispeed

   accept:      termbits.h defined value for speed.
   perform:     set the termios structure output speed value.
   return:      0 on success.
                -1 on failure.
   exceptions:  none.
   comments:    the termios structure is not what the kernel is using.  use
                tcsetattr to get the kernel to read the speed after setting
                it here.
                A value of zero means input=output speed.

                low level speed control
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_cfsetispeed (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jint speed
   )
{
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	return( cfsetispeed ( &cp[PortNumber].ttyset, (int) speed));
#else
#endif /* newer libc in linux */
#ifdef __FreeBSD__
	return( cfsetspeed ( &cp[PortNumber].ttyset, (int) speed));
#endif
}
/*----------------------------------------------------------
SerialInputStream_GetSpeed

   accept:     none.
   perform:    get the speed in bits-per-second.
   return:     the speed as defined in termbits.h 
   exceptions: none.
   comments:   high level speed control.  called from
                  RXTXPort.java
                  SerialInputStream.java
                  Test.java
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_GetSpeed (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("GetSpeed", PortNumber);
	return(speed_t2bps(cfgetispeed(&Port->ttyset)));
}
/*----------------------------------------------------------
SerialInputStream_SetSpeed

   accept:     speed.  See bps2speed_t() for valid speeds.  
   perform:    set the termios structure speed to speed.
   return:     none.
   exceptions: If fd is <1 bad file descriptor
               tcgetattr tcgetspeed, tcsetspeed and tcsetattr failures. 
   comments:   high level speed control.  called from
                  RXTXPort.java
                  SerialInputStream.java
                  Test.java
----------------------------------------------------------*/ 

JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_SetSpeed (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
 	jint speed
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("SetSpeed", PortNumber);
	if (Port->fd < 1) { errno= EBADFD;goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	if (bps2speed_t(speed)<=0) goto fail;
#ifdef __FreeBSD__
	if(cfsetspeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	if(cfsetispeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail; 
	return;
fail:
	IOException(Env,"SetSpeed", strerror(errno));
	
}
/*----------------------------------------------------------
Native_ReadBytes

   accept:      int   fd       file descriptor to read from
                jbyte *body    array to fill after reading
                int   offset   number of bytes to discard 
                               before read
                int   length   number of bytes to read
   perform:     read, fill array, return status. 
   return:      status of read
                -1 fail (ioexception)
                 0 timeout
                >0 number of bytes read
   exceptions:  none.  
   comments:    work in progress.  expect problems.
----------------------------------------------------------*/ 
int Native_ReadBytes(int fd,jbyte *body,int offset,int  length,int PortNumber)
{
	int ret=0,i;
	struct commport *Port=&cp[PortNumber];
	unsigned char *buffer= (unsigned char *)malloc(length);
#ifdef WAYNE
	struct timeval start, now, result, sleep;
	long int current;
	int base, newlen;

	if(buffer==0) goto fail;
	toss_debug("Native_ReadBytes", PortNumber);
/*----------------------------------------------------------------------
The Following code was contributed by Wayne Roberts.   As Wayne has this
code working in some demanding situations its worthy of trying to get in.
----------------------------------------------------------------------*/

	if(fd>0){
		gettimeofday(&start, NULL);
		sleep.tv_sec = 0;
		/* sleep ~1mS per byte for 9600, ~1/2mS for 19200, etc.. */
		sleep.tv_usec = (1.0/Port->BAUDRATE)*1E+7*length; 
		/* wait time for uart to fill */
		/* select(0, NULL, NULL, NULL, &sleep);	 */
		ret = read(fd, buffer, length);
		gettimeofday(&now, NULL);
		timeval_subtract(&result, &now, &start);
		current = (result.tv_sec*1000000)+result.tv_usec;
		while (ret < length && 
			current < Port->JTIMEOUT && 
			ret < Port->THRESHOLD) 
		{
			base = ret;
			newlen = length-ret;
			ret = read(fd, &buffer[base], newlen);
			ret += base;
/*			if (Port->no_block != BLOCKING) { */
				gettimeofday(&now, NULL);
				timeval_subtract(&result, &now, &start);
				current = (result.tv_sec*1000000)+result.tv_usec;
/*			} */
		}
	}
	for(i=0;i<ret;i++) body[i]=buffer[i];
	free(buffer); 
	return(ret);
#else
	int ReadFlag=-1;
	if(buffer==0) goto fail;
	toss_debug("Native_ReadBytes", PortNumber);
	if(fd>0){
		/* determine number of bytes */
		ioctl (fd, FIONREAD, &ret);
		if (Port->no_block == BLOCKING && ret == 0) 
			ret = 1; /*wait on 1 byte */
		if (ret > 0) {
			/* limit to buffer size */
			if (ret > length) ret = length; 
			ReadFlag=read(fd, buffer, ret);
			for(i=0;i<length;i++) body[i]=buffer[i];
		}
		free(buffer); 
		return(ReadFlag);
	} 
#endif
fail:
	free(buffer); 
	return(0);
}

/*----------------------------------------------------------
SerialInputStream_nativeread

   accept:      none.
   perform:     Read a byte in a blocking or or nonblocking fashion.
   return:      The byte read
                -1 if read fails.
   exceptions:  if read fails.
   comments:    behaviour is not specified if asked to read > SSIZE_MAX
                read is not settled.  __read() was used to bypass green
                threads.  linux uses native threads now.  __read() is known
                to not work with all libs (bsd for instance).  I'm not sure
                what is best and portable.
----------------------------------------------------------*/ 

jint nativeread ( JNIEnv *Env, jobject jobj, jint PortNumber)
{return(Java_gnu_io_SerialInputStream_nativeread(Env,jobj,PortNumber));}

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_nativeread (
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber
	)
{ 
	int ReadFlag;
	unsigned char buf[1]; /* unsigned char for binary? */
	struct commport *Port=&cp[PortNumber];

	toss_debug_blocking("read", PortNumber);
	if (Port->fd < 0) {errno=EBADFD ;goto fail; }
	ReadFlag= Native_ReadBytes(Port->fd, buf,0, 1, PortNumber); 
	if(ReadFlag<0) goto fail;
	else if(ReadFlag==0) { 
#ifdef DEBUG
		printf("%s\n",strerror(ETIME));
#endif
		return(-1); 
	} 
	return((jint) buf[0]);
fail:
	IOException(Env,"read", strerror(errno));
	return(-1);
}
/*----------------------------------------------------------
SerialInputStream_readBytes

   accept:       offset (bytes to skip) and Length (bytes to read)
   perform:      skip offset bytes and read Length bytes in a blocking
                 or nonblocking fashion.
   return:       bytes read on success
                 -1 on read error.
   exceptions:   read error
   comments:     behaviour is not specified if asked to read > SSIZE_MAX
                 FIXME offset is not used.
                 Offset?
----------------------------------------------------------*/ 

jint readBytes (JNIEnv *Env,jobject jobj,jint PortNumber,jbyteArray jbarray, jint Offset, jint Length)
{return(Java_gnu_io_SerialInputStream_readBytes(Env,jobj,PortNumber,jbarray,Offset,Length));}

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_readBytes (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jbyteArray jbarray, 
	jint Offset,
	jint length
	)
{  
	struct commport *Port=&cp[PortNumber];
	jbyte *body = GetByteArrayElements(Env, jbarray, 0);
	int ReadFlag;

	toss_debug_blocking("readBytes", PortNumber);

	/* should really malloc buf   FIXME */
	if (Port->fd < 0) { errno=EBADFD;goto fail; }
        ReadFlag= Native_ReadBytes(Port->fd, body,0, length, PortNumber);
        if(ReadFlag<0) goto fail;
	else if(ReadFlag==0) { 
		printf("%s\n",strerror(ETIME));
		return(-1); 
	} 
	ReleaseByteArrayElements(Env, jbarray, body, 0);
	return(ReadFlag);
fail:
	ReleaseByteArrayElements(Env, jbarray, body, 0);
	IOException(Env,"readBytes", strerror(errno));
	return(-1);

}
/*----------------------------------------------------------
SerialInputStream_skip

   accept:      jl number of bytes to attempt to skip.
   perform:     toss out jl bytes from the input stream.
   return:      number of bytes actually skipped on success.
                -1 on read failure.
   exceptions:  on read failure
   comments:    Politely reverts to a slower algorithm if 
                there is no memory for a buffer.
----------------------------------------------------------*/ 

JNIEXPORT jlong JNICALL Java_gnu_io_SerialInputStream_skip (
	JNIEnv *Env, 
	jobject jobj, 
	jint PortNumber,
	jlong jl
   )
{ 
	jint bytes_read;
	jlong skipped = 0;

	/*
	  If there is no memory available for a buffer, do it the 
	  slow way.
	*/
	jbyteArray jbarray = malloc( sizeof( jbyte ) * MAX_SKIP_BUFFER_SIZE );
	toss_debug("skip", PortNumber);
	if( jbarray == NULL ) { 
		for( skipped = 0 ; skipped < jl ; skipped++ ) 
			if(nativeread( Env, jobj, PortNumber ) == -1 ) break;
      		return skipped;
	}

	/*
	  Read into the buffer as many times as necessary to add up to
	  the amount to skip.
	*/
	while( skipped < jl ) {
		bytes_read = readBytes(Env, 
					jobj, 
					PortNumber, 
					jbarray,
					0, 
					MAX_SKIP_BUFFER_SIZE );
		if( bytes_read == -1 ) break;
		skipped += bytes_read;	
	}
	free( jbarray );
	return skipped;
}

/*----------------------------------------------------------
SerialInputStream_nativeavailable
   accept:      none.
   perform:     find out if bytes are available for reading.
   return:      available bytes.
   exceptions:  tcgettattr fail and tcsetattr fail.
   comments:    The thread RXMonitor will use this function to instigate hwfc.
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_nativeavailable (
	JNIEnv *Env, 
	jobject jobj,
	jint PortNumber
   )
{
	int result;
	struct commport *Port=&cp[PortNumber];

	toss_debug("available", PortNumber);
 	/* check for available bytes  */
	if(Port->fd_is_needed||Port->fd < 0) return(JNI_FALSE);
	if(ioctl (Port->fd, FIONREAD, &result)==0)
		return ((jint) result);
	return(-1);
}

/*----------------------------------------------------------
SerialInputStream_setHWFC

   accept:      state.  (JNI_FALSE 0, JNI_TRUE 1)
   perform:     set hardware flow control to state.  This is used by
                InputSerialStream_nativeavailable() (was used...)
   return:      none.
   exceptions:  Bad FileDescriptor, tcsetattr fail.
   comments:    setHWFC has been implemented to signify that the available
                function should also handle hardware flow control Default
                state is no flow control.  It may be required to toggle
                CLOCAL in c_cflag.
----------------------------------------------------------*/

JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_setHWFC (
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
        jint state
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("setHWFC", PortNumber);
	if (Port->fd < 0) {  errno=EBADFD;goto fail; }
        Port->hardwarefc = state;
	/* set flag based on state */
	if (state == JNI_TRUE)
   		Port->ttyset.c_cflag |= HARDWARE_FLOW_CONTROL;
	else
   		Port->ttyset.c_cflag &= ~HARDWARE_FLOW_CONTROL;

	/* update attributes */
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail;
	return;
fail:
	IOException(Env,"setHWFC(hw)", strerror(errno));
}

/*----------------------------------------------------------
SerialInputStream_noblock

   accept:      state.  (BLOCKING 0, NONBLOCK 1)
   perform:     set the blocking state.  reading functions check the state
                and behave accordingly.
   return:      none.
   exceptions:  none.
   comments:    noblock has been implemented to signify that nonblocking
                operation is required. This is necessary because the port
                always blocks!!!  Default state is blocking.
----------------------------------------------------------*/

JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_noblock (
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
        jint state
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("noblock", PortNumber);
        Port->no_block = state;
}

/*----------------------------------------------------------
SerialOutputStream_setoutputbuffersize(int size);

   accept:      size.
   perform:     Set the buffer size for writes.
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialOutputStream_setoutputbuffersize(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
	jint size
)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("setoutputbuffersize", PortNumber);
	Port->OutputBufferSize=size;
}
/*----------------------------------------------------------
SerialOutputStream_getoutputbuffersize();

   accept:      none.
   perform:     return the size of OutputBufferSize
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jint JNICALL Java_gnu_io_SerialOutputStream_getoutputbuffersize(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("getoutputbuffersize", PortNumber);
	return(Port->OutputBufferSize);
}
/*----------------------------------------------------------
SerialInputStream_setinputbuffersize

   accept:      size.
   perform:     
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_setinputbuffersize(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
	jint size
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("setinputbuffersize", PortNumber);
	Port->InputBufferSize=size;
}
/*----------------------------------------------------------
SerialInputStream_getinputbuffersize

   accept:      none.
   perform:     
   return:      inputbuffersize.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_getinputbuffersize(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("getinputbuffersize", PortNumber);
	return(Port->InputBufferSize);
}
/*----------------------------------------------------------
SerialInputStream_enablereceiveframing

   accept:      Byte.
   perform:     
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_enablereceiveframing(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
	jint Byte
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("enablereceiveframing", PortNumber);
	Port->FrameByte=Byte;
}
/*----------------------------------------------------------
SerialInputStream_disablereceiveframing

   accept:      none.
   perform:     
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_disablereceiveframing(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("disablereceiveframing", PortNumber);
	Port->FrameByte=0;
}
/*----------------------------------------------------------
SerialInputStream_getreceiveframingbyte

   accept:      none.
   perform:     
   return:      receiveframingbyte.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_getreceiveframingbyte(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("getreceiveframingbyte", PortNumber);
	return(Port->FrameByte);
}
/*----------------------------------------------------------
SerialInputStream_isreceiveframingenabled

   accept:      none.
   perform:     
   return:      True if framing is enabled.  otherwise false.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isreceiveframingenabled(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("isreceiveframingenabled", PortNumber);
	if (Port->FrameByte!=0)
		   return(JNI_TRUE);
	else return(JNI_FALSE);
}
/*----------------------------------------------------------
SerialInputStream_gettimeout

   accept:      none.
   perform:     
   return:      timeout.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_gettimeout(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("gettimeout", PortNumber);
#ifdef WAYNE
	return(Port->JTIMEOUT/1000);
#else
	if (Port->fd < 0) { errno=EBADFD; goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	return(Port->ttyset.c_cc[VTIME]*100);
fail:
	IOException(Env,"gettimeout", strerror(errno));
	return(-1);
#endif
}
/*----------------------------------------------------------
SerialInputStream_settimeout

   accept:      time.  The number of milliseconds (0.001) to wait on reads.
   perform:     Change the VTIME value of the termios structure 
   return:      none.
   exceptions:  none.
   comments:    time is sent in seconds.  VTIME is in deciseconds (0.1)
                increments.  VTIME only works in nonconanical mode.
                VTIME and VMIN combine to determine when reads return.
                Obviously we lose some resolution.

                Perhaps a better implementation would involve usleep().  We
                can get close to usec resolution with that.  
                bah.  compile, release, wait.
----------------------------------------------------------*/

JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_settimeout (
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
        jint time
   )
{
#ifndef WAYNE
#endif
	struct commport *Port=&cp[PortNumber];

	toss_debug("settimeout", PortNumber);

#ifdef WAYNE
	Port->JTIMEOUT=time*1000;
	return;
#else
	if (Port->fd < 0) { errno=EBADFD;goto fail;}
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	Port->ttyset.c_cc[VTIME]=(int) time/100;
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail;
	return;
fail:
	IOException(Env,"settimeout", strerror(errno));
#endif
}

/*----------------------------------------------------------
SerialInputStream_enablereceivethreshold

   accept:      time.
   perform:     Set the receive threshold (number of bytes read before
                immediate return)
   return:      none.
   exceptions:  none.
   comments:    This is for blocking reads.
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_enablereceivethreshold(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber,
	jint size
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("enablereceivethresho", PortNumber);
#ifdef WAYNE
	Port->THRESHOLD = size;
	return;
#else
	if (Port->fd < 0) { errno=EBADFD; goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	Port->ttyset.c_cc[VMIN]=size;
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail;
	return;
fail:
	IOException(Env,"enbablereceivethreshold", strerror(errno));
	return;
#endif
}
/*----------------------------------------------------------
SerialInputStream_getreceivethreshold

   accept:      none.
   perform:     none.    
   return:      ReadThreshold.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jint JNICALL Java_gnu_io_SerialInputStream_getreceivethreshold(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("getreceivethreshold", PortNumber);
#ifdef WAYNE
	return(Port->THRESHOLD);
#else
	if (Port->fd < 0) { errno=EBADFD;goto fail;}
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	return(Port->ttyset.c_cc[VMIN]);
fail:
	IOException(Env,"getreceivethreshold", strerror(errno));
	return(-1);
#endif
}
/*----------------------------------------------------------
SerialInputStream_disablereceivethreshold

   accept:      none.
   perform:     
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT void JNICALL Java_gnu_io_SerialInputStream_disablereceivethreshold(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("disablereceivethreshold", PortNumber);
#ifdef WAYNE
	Port->THRESHOLD = 0;
#else
	if (Port->fd < 0) { errno=EBADFD;goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	Port->ttyset.c_cc[VMIN]=0;
	if (tcsetattr(Port->fd, TCSAFLUSH, &Port->ttyset) <0) goto fail;
	return;
fail:
	IOException(Env,"disablereceivethreshold", strerror(errno));
	return;
#endif
}
/*----------------------------------------------------------
SerialInputStream_isreceivethresholdenabled

   accept:      none.
   perform:     return TRUE if threshold is >0 otherwise return FALSE
   return:      none.
   exceptions:  none.
   comments:    
----------------------------------------------------------*/
JNIEXPORT jboolean JNICALL Java_gnu_io_SerialInputStream_isreceivethresholdenabled(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
	)
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("isreceivethresholdenabled", PortNumber);
#ifdef WAYNE
        if (Port->THRESHOLD > 0)
                return(JNI_TRUE);
        else
                return(JNI_FALSE);
#else
	if (Port->fd < 0) { errno=EBADFD;goto fail; }
	if(tcgetattr(Port->fd, &Port->ttyset)< 0) goto fail;
	if(Port->ttyset.c_cc[VMIN]==0) return(JNI_FALSE);
	else return(JNI_TRUE);
fail:
	IOException(Env,"isreceivethresholdenabled", strerror(errno));
	return(JNI_FALSE);
#endif
}

/* MISC/Misc FUNCTIONS */
/*----------------------------------------------------------
 signal_handler

   accept:      
   perform:     
   return:      
   exceptions:  
----------------------------------------------------------*/ 
void bonk(int PortNumber, int a, int b,char *name){
	if(a!=b){
#ifdef DEBUG
		printf("%s= +%i ",name, a-b);
#endif
		sIGiO=PortNumber;
	}
}
void signal_handler (int status)
{ 
#ifdef ASYNC
	unsigned int result;
	int i;

        if (status != SIGIO) {
                fprintf(stderr, "unknown signal:%d\n", status);
                return;
        }
#ifdef REPORTEMPTY
	else if (status == SIGCONT)
		printf(" <transmit_buffer_empty>\n");
#endif
#ifdef DEBUG
	for(i=0;i<4;i++){
		printf("port=%i fd=%i available=%i numberopen=%i\n",i,cp[i].fd,cp[i].available,cp[i].number_of_times_opened);
	}
#endif
	for(i=0;i<4;i++){
		if(cp[i].fd > 0){
        		ioctl (cp[i].fd, TIOCMGET, &result);
#ifdef EVENTNOTIFICATION
#if defined(__linux__) /* this will change */
			cp[i].osis=cp[i].sis;
        		ioctl(cp[i].fd, TIOCGICOUNT,&cp[i].sis);
#ifdef DEBUG
			printf("          \\|/ ____ \\|/   ");
			printf("status=%i ",status);
#endif
#ifdef FULL_EVENT /* errors, rx/tx are not available on 2.0.36 linux kernels */
			bonk(i,cp[i].sis.rx,cp[i].osis.rx,"rx");
			bonk(i,cp[i].sis.tx,cp[i].osis.tx,"tx");
			bonk(i,cp[i].sis.frame,cp[i].osis.frame,"frame");
			bonk(i,cp[i].sis.overrun,cp[i].osis.overrun,"overrun");
/*
			bonk(i,cp[i].sis.outputempty,cp[i].osis.outputempty,"outputempty");
*/
			bonk(i,cp[i].sis.parity,cp[i].osis.parity,"parity");
			bonk(i,cp[i].sis.brk,cp[i].osis.brk,"brk");
#endif /* FULL_EVENT */
#ifdef DEBUG
			printf("\n");
			printf("           @~/ Oo \\~@   ");
#endif
			bonk(i,cp[i].sis.cts,cp[i].osis.cts,"cts");
			bonk(i,cp[i].sis.dsr,cp[i].osis.dsr,"dsr");
			bonk(i,cp[i].sis.rng,cp[i].osis.rng,"rng");
			bonk(i,cp[i].sis.dcd,cp[i].osis.dcd,"dcd");
#ifdef DEBUG
			printf("\n");
			printf(" sIGiO    /_( \\__/ )_\\   ");
#endif
#ifdef DEBUG
			printf("\n");
			printf("             \\__U_/      %s %i ",cp[i].name,i);
#endif
#endif /* __linux__ */
#endif /* EVENTNOTIFICATION */
#ifdef DEBUG
        		if(result & TIOCM_CTS ) printf("CTS ");
        		if(result & TIOCM_DSR ) printf("DSR ");
        		if(result & TIOCM_CD ) printf("CD ");
        		if(result & TIOCM_RI ) printf("RI ");
        		if(result & TIOCM_RTS ) printf("RTS ");
        		if(result & TIOCM_DTR ) printf("DTR ");
			printf("\n");
#endif
		}
	}
#endif /* ASYNC */
}
/*----------------------------------------------------------
 timeval_subtract

   accept:      result, x and y time values 
   perform:     result=y - x
   return:      result.
   exceptions:  none.
   comments:    Wayne Roberts read implementation uses this.
                timeval subtract function is from glibc info page
----------------------------------------------------------*/ 
int timeval_subtract(result, x, y)
        struct timeval *result, *x, *y; {
        /* perform carry for the later subtraction by updating Y */
        if (x->tv_usec < y->tv_usec) {
                int nsec = (y->tv_usec - x->tv_usec)/1000000+1;
                y->tv_usec -= 1000000*nsec;
                y->tv_sec += nsec;
	}
        if (x->tv_usec - y->tv_usec > 1000000) {
                int nsec = (y->tv_usec - x->tv_usec) / 1000000;
                y->tv_usec += 1000000 * nsec;
                y->tv_sec -= nsec;
	}

        /* compute the time remaining to wait, tv_usec is certainly positive */
        result->tv_sec = x->tv_sec - y->tv_sec;
        result->tv_usec = x->tv_usec - y->tv_usec;

        /* return 1 if result is negative */
        return x->tv_sec < y->tv_sec;
}

/*----------------------------------------------------------
show_ctl

   accept:      fd file descriptor. 
   perform:     Display the control settings.
   return:      none.
   exceptions:  none.
   comments:    This appears to be a legitimate use of ioctl.
----------------------------------------------------------*/ 

void show_ctl(int fd, struct commport *cp) {
	unsigned int result;
	ioctl (fd, TIOCMGET, &result);

	if (result == cp->last_TIOCMGET_result) return;
	
	cp->last_TIOCMGET_result = result;
#ifdef TIOCM_LE
	if (result & TIOCM_LE) printf("TIOCM_LE\n");
#endif
#ifdef TIOCM_DTR
	if (result & TIOCM_DTR) printf("TIOCM_DTR\n");
#endif
#ifdef TIOCM_RTS
	if (result & TIOCM_RTS) printf("TIOCM_RTS\n");
#endif
#ifdef TIOCM_ST
	if (result & TIOCM_ST) printf("TIOCM_ST\n");
#endif
#ifdef TIOCM_SR
	if (result & TIOCM_SR) printf("TIOCM_SR\n");
#endif
#ifdef TIOCM_CTS
	if (result & TIOCM_CTS) printf("TIOCM_CTS\n");
#endif
#ifdef TIOCM_CAR
	if (result & TIOCM_CAR) printf("TIOCM_CAR, CD\n");
#endif
#ifdef TIOCM_RNG
	if (result & TIOCM_RNG) printf("TIOCM_RNG, RI\n");
#endif
#ifdef TIOCM_DSR
	if (result & TIOCM_DSR) printf("TIOCM_DSR\n");
#endif

}

   
/*----------------------------------------------------------
 makeraw
   accept:     *ttyset (pointer to the termios structure) 
   perform:     set the termios into a sane condition.
   return:      none.
   exceptions:  none.
   comments:    update SerialInput/OutputStream_cfmakeraw() comments if
                this is changed.
                changing the termios structer changes no behavior unless
                tcsetattr is called afterwards.
                -implementation of cfmakeraw from linux termios
----------------------------------------------------------*/

void makeraw(struct termios * tty){

#ifdef GARBAGE
	tty->c_iflag = IXON | IXOFF | IGNBRK | IGNPAR |ISTRIP;
	tty->c_lflag = 0;
	tty->c_oflag = 0;
	tty->c_cflag |= CREAD |HUPCL | CS8; /* CLOCAL can be added to 
                                        ignore modem control lines */
	tty->c_cc[VMIN] = 0;
	tty->c_cc[VTIME] = 1;
#endif
	tty->c_iflag = IGNPAR;
	tty->c_lflag = 0;
	tty->c_oflag = 0;
	tty->c_cflag |= CREAD | CS8; /* CLOCAL can be added to 
                                        ignore modem control lines */
	tty->c_cc[VMIN] = 0;
	tty->c_cc[VTIME] = 1;
}

/*----------------------------------------------------------
 bps2speed_t 
   accept:     val - speed in bits-per-second
   perform:    convert bits-per-second to a speed_t constant
   return:     speed_t constant on success
               -1 on fail
   exceptions: none.
   comments:   Only the lowest level code should know about
               the magic constants.
----------------------------------------------------------*/ 

int bps2speed_t(int val)
{
	switch(val){
		case 0:     return(B0);
		case 50:    return(B50);
		case 75:    return(B75);
		case 110:   return(B110);
		case 134:   return(B134);
		case 150:   return(B150);
		case 200:   return(B200);
		case 300:   return(B300);
		case 600:   return(B600);
		case 1200:  return(B1200);
		case 1800:  return(B1800);
		case 2400:  return(B2400);
		case 4800:  return(B4800);
		case 9600:  return(B9600);
		case 19200: return(B19200);
		case 38400: return(B38400);
#ifdef B57600
		case 57600: return(B57600);
#endif
#ifdef B115200
		case 115200: return(B115200);
#endif
#ifdef B230400
		case 230400: return(B230400);
#endif
#ifdef B460800
		case 460800: return(B460800);
#endif
		default:    return(-1);
	}
}

/*----------------------------------------------------------
 speed_t2bps 
   accept:     val - integer speed_t constant
   perform:    convert a speed_t constant to bits-per-second.
   return:     bits-per-second value on success
               -1 on fail
   exceptions: none.
   comments:   none.
----------------------------------------------------------*/ 

int speed_t2bps(int val)
{
	switch(val){
		case B0:     return(0);
		case B50:    return(50);
		case B75:    return(75);
		case B110:   return(110);
		case B134:   return(134);
		case B150:   return(150);
		case B200:   return(200);
		case B300:   return(300);
		case B600:   return(600);
		case B1200:  return(1200);
		case B1800:  return(1800);
		case B2400:  return(2400);
		case B4800:  return(4800);
		case B9600:  return(9600);
		case B19200: return(19200);
		case B38400: return(38400);
#ifdef B57600
		case B57600: return(57600);
#endif
#ifdef B115200
		case B115200: return(115200);
#endif
#ifdef B230400
		case B230400: return(230400);
#endif
#ifdef B460800
		case B460800: return(460800);
#endif
		default:    return(-1);
	}
}

void toss_error(JNIEnv *Env,char *msg1, char *msg2){
#ifdef DEBUG
		IOException(Env,msg1, msg2);
#endif
}
void toss_debug_blocking(char *function_name, int PortNumber){
#ifdef DEBUG
	printf("%-20s pid=%i PortName=%10s PortNumber=%i count=%i Block= %i\n",
		function_name,
		getpid(),
		cp[PortNumber].name,
		PortNumber,
		cp[PortNumber].number_of_times_opened,
		cp[PortNumber].no_block
		);
#endif
}
void toss_debug(char *function_name,int PortNumber){
#ifdef DEBUG
	printf("%-20s pid=%i PortName=%10s PortNumber=%i count=%i fd=%i\n",
		function_name, 
		getpid(),
		cp[PortNumber].name,
		PortNumber,
		cp[PortNumber].number_of_times_opened,
		cp[PortNumber].fd
		);
#endif
}
/*----------------------------------------------------------
IOException
   Throw an IOException.  I'm assuming most the error handling
   is done in java.  most functions do not handle the errors, only
   report them. 

   accept:      Env (keyhole to java) *function (string describing function)
                *msg (error message)
   perform:     Throw a java/io/IOException and print the function/message.
   return:      none. 
   exceptions:  haha!
   comments:
----------------------------------------------------------*/ 
void IOException( JNIEnv *Env, char *function, char *msg)
{
	jclass ECls;

#ifndef NO_VERBOSE_IOEXCEPTIONS
   printf("\nSerialImp.c is throwing an exception from  Java_gnu_io_Serial(In/Out)put_stream_%s\n----------\nReported error was: %s\n----------\n",function,msg);
#endif /* NO_VERBOSE_IOEXCEPTIONS */
	ECls = FindClass(Env, "java/io/IOException");
	if (ECls == 0) return;
	ThrowNew(Env, ECls, msg);
}

/*----------------------------------------------------------
Termios_strlenccc

   accept:     
   perform:    get the string length of termios.c_cc
   return:     length of termios.c_cc
   exceptions: none.
   comments:   dont think this was ever used properly/finished.
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_Termios_strlenccc(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
   )
{
	struct commport *Port=&cp[PortNumber];

	toss_debug("strlenccc",PortNumber);
	return(strlen(Port->ttyset.c_cc));
}
/*----------------------------------------------------------
Termios_ShowNativeTermios


   accept:     
   perform:    print the termios structure stored on the native side
   return:     
   return:     0 on success
               -1 on fail
   exceptions: none.
   comments:  
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_Termios_ShowNativeTermios(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
   )
{
	struct termios *ttyset=&cp[PortNumber].ttyset;

	printf("----- native ----->\n");
#if defined(__linux__)   /* c_line not on all systems */
	printf("c_line = %c", ttyset->c_line);
#endif   /* c_line not on all systems */
	translate_c_cc(ttyset->c_cc);
	translate_ciflag((int) ttyset->c_iflag);
	translate_coflag((int) ttyset->c_oflag);
	translate_ccflag((int) ttyset->c_cflag);
	translate_clflag((int) ttyset->c_lflag);
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	printf("c_ispeed = %i ", (int) ttyset->c_ispeed);
	printf("c_ospeed = %i\n", (int) ttyset->c_ospeed);
#else
	printf("cfgetospeed = ");
	translate_speed(cfgetospeed(ttyset));
	printf("\ncfgetispeed = ");
	translate_speed(cfgetispeed(ttyset));
	printf("\n");
#endif /* newer libc in linux */
	printf("----- native ----->\n");
	return(0);
}

/*----------------------------------------------------------
translate_ciflag  
   accept:     c_iflag.  (a part of the termios structure)
   perform:    translate and print attributes
   return:     none.
   exceptions: none.
   comments:
----------------------------------------------------------*/ 
void translate_ciflag(int c_iflag) 
{
	/* this is simple but ugly */
	printf("c_iflag: ");
	if (c_iflag & IGNBRK) printf("IGNBRK ");
	if (c_iflag & BRKINT) printf("BRKINT ");
	if (c_iflag & IGNPAR) printf("IGNPAR ");
	if (c_iflag & PARMRK) printf("PARMRK ");
	if (c_iflag & INPCK) printf("INPCK ");
	if (c_iflag & ISTRIP) printf("ISTRIP ");
	if (c_iflag & INLCR) printf("INLCR ");
	if (c_iflag & IGNCR) printf("IGNCR ");
	if (c_iflag & ICRNL) printf("ICRNL ");
#ifdef IUCLC /* legacy support */
	if (c_iflag & IUCLC) printf("IUCLC ");
#endif
	if (c_iflag & IXON) printf("IXON ");
	if (c_iflag & IXANY) printf("IXANY ");
	if (c_iflag & IXOFF) printf("IXOFF ");
	if (c_iflag & IMAXBEL) printf("IMAXBEL ");
	printf("\n");
}

/*----------------------------------------------------------
translate_coflag  
   accept:     coflag (a member of the termios structure)
   perform:    translate and print attributes
   return:     none.
   exceptions: none.
   comments:   
----------------------------------------------------------*/ 
void translate_coflag(int c_oflag) 
{
	/* this is simple but ugly */
	printf("c_oflag: ");
	if (c_oflag & OPOST) printf("OPOST ");
#ifdef OLCUC /* legacy support */
	if (c_oflag & OLCUC) printf("OLCUC ");
#endif
	if (c_oflag & ONLCR) printf("ONLCR ");
#ifdef OCRNL
	if (c_oflag & OCRNL) printf("OCRNL ");
#endif
#ifdef ONOCR
	if (c_oflag & ONOCR) printf("ONOCR ");
#endif
#ifdef ONLRET
	if (c_oflag & ONLRET) printf("ONLRET ");
#endif
#ifdef OFILL
	if (c_oflag & OFILL) printf("OFILL ");
#endif
#ifdef OFDEL
	if (c_oflag & OFDEL) printf("OFDEL ");
#endif
#ifdef ONLDLY
	if (c_oflag & NLDLY) printf("NLDLY ");
#endif
#ifdef OCRDLY
	if (c_oflag & CRDLY) printf("CRDLY ");
#endif
#ifdef TABDLY
	if (c_oflag & TABDLY) printf("TABDLY ");
#endif
#ifdef BSDLY
	if (c_oflag & BSDLY) printf("BSDLY ");
#endif
#ifdef VTDLY
	if (c_oflag & VTDLY) printf("VTDLY ");
#endif
#ifdef FFDLY
	if (c_oflag & FFDLY) printf("FFDLY ");
#endif
#ifdef OXTABS
	if (c_oflag & OXTABS) printf("OXTABS ");
#endif
#ifdef ONOEOT
	if (c_oflag & ONOEOT) printf("ONOEOT ");
#endif
	printf("\n");
}

/*----------------------------------------------------------
translate_ccflag  
   accept:     ccflag (a member of the termios structure)
   perform:    translate and print attributes
   return:     none.
   exceptions: none.
   comments:
----------------------------------------------------------*/ 
void translate_ccflag(int c_cflag) 
{
	/* this is simple but ugly */
	printf("c_cflag: ");
	if (c_cflag & CSIZE) {
		printf("CSIZE ");
		if ((c_cflag & CSIZE) == CS5) printf("CS5 ");
		if ((c_cflag & CSIZE) == CS6) printf("CS6 ");
		if ((c_cflag & CSIZE) == CS7) printf("CS7 ");
		if ((c_cflag & CSIZE) == CS8) printf("CS8 ");
	}
	if (c_cflag & CSTOPB) printf("CSTOPB ");
	if (c_cflag & CREAD) printf("CREAD ");
	if (c_cflag & PARENB) printf("PARENB ");
	if (c_cflag & PARODD) printf("PARODD ");
	if (c_cflag & HUPCL) printf("HUPCL ");
	if (c_cflag & CLOCAL) printf("CLOCAL (Probably really bad)");
	if (c_cflag & HARDWARE_FLOW_CONTROL) printf("HARDWARE_FLOW_CONTROL ");
#ifdef CCTS_OFLOW
	if (c_cflag & CCTS_OFLOW) printf("CCTS_OFLOW ");
#endif
#ifdef CCTS_IFLOW
	if (c_cflag & CCTS_IFLOW) printf("CCTS_IFLOW ");
#endif
#ifdef MDMBUF
	if (c_cflag & MDMBUF) printf("MDMBUF ");
#endif
	printf("\n");
}
/*----------------------------------------------------------
translate_clflag  
   accept:     c_lflag (a member of the termios structure)
   perform:    translate and print attributes
   return:     none.
   exceptions: none.
   comments:
----------------------------------------------------------*/ 
void translate_clflag(int c_lflag) 
{
	/* this is simple but ugly */
	printf("c_lflag: ");
	if (c_lflag & ISIG) printf("ISIG ");
	if (c_lflag & ICANON) printf("ICANON ");
	if (c_lflag & ECHO) printf("ECHO ");
	if (c_lflag & ECHOE) printf("ECHOE ");
	if (c_lflag & ECHOK) printf("ECHOK ");
	if (c_lflag & ECHONL) printf("ECHONL ");
#ifdef ECHOCTL
	if (c_lflag & ECHOCTL) printf("ECHOCTL ");
#endif
#ifdef ECHOPRT
	if (c_lflag & ECHOPRT) printf("ECHOPRT ");
#endif
#ifdef ECHOKE
	if (c_lflag & ECHOKE) printf("ECHOKE ");
#endif
#ifdef FLUSHO
	if (c_lflag & FLUSHO) printf("FLUSHO ");
#endif
	if (c_lflag & NOFLSH) printf("NOFLSH ");
	if (c_lflag & TOSTOP) printf("TOSTOP ");
#ifdef PENDING
	if (c_lflag & PENDIN) printf("PENDIN ");
#endif
	if (c_lflag & IEXTEN) printf("IEXTEN ");
#ifdef XCASE /* legacy support */
	if (c_lflag & XCASE) printf("XCASE ");
#endif
#ifdef ALTWERASE
	if (c_lflag & ALTWERASE) printf("ALTWERASE ");
#endif
#ifdef EXTPROC
	if (c_lflag & EXTPROC) printf("EXTPROC ");
#endif
#ifdef NOKERNINFO
	if (c_lflag & NOKERNINFO) printf("NOKERNINFO ");
#endif
	printf("\n");
}

/*----------------------------------------------------------
translate_c_cc
   accept:     c_cc.  (a part of the termios structure)
   perform:    translate and print attributes
   return:     none.
   exceptions: none.
   comments:
----------------------------------------------------------*/ 
void translate_c_cc(char *c_cc)
{
	if(c_cc[VINTR]) printf("c_cc[VINTR] = %i\n",c_cc[VINTR]);
	if(c_cc[VQUIT]) printf("c_cc[VQUIT] = %i\n",c_cc[VQUIT]);
	if(c_cc[VERASE]) printf("c_cc[VERASE] = %i\n",c_cc[VERASE]);
	if(c_cc[VKILL]) printf("c_cc[VKILL] = %i\n",c_cc[VKILL]);
	if(c_cc[VEOF]) printf("c_cc[VEOF] = %i\n",c_cc[VEOF]);
	if(c_cc[VTIME]) printf("c_cc[VTIME] = %i\n",c_cc[VTIME]);
	if(c_cc[VMIN]) printf("c_cc[VMIN] = %i\n",c_cc[VMIN]);
#ifdef VSWTC
	if(c_cc[VSWTC]) printf("c_cc[VSWTC] = %i\n",c_cc[VSWTC]);
#endif
	if(c_cc[VSTART]) printf("c_cc[VSTART] = %i\n",c_cc[VSTART]);
	if(c_cc[VSTOP]) printf("c_cc[VSTOP] = %i\n",c_cc[VSTOP] );
	if(c_cc[VSUSP]) printf("c_cc[VSUSP] = %i\n",c_cc[VSUSP]);
	if(c_cc[VEOL]) printf("c_cc[VEOL] = %i\n",c_cc[VEOL]);
	if(c_cc[VREPRINT]) printf("c_cc[VREPRINT] = %i\n",c_cc[VREPRINT]);
	if(c_cc[VDISCARD]) printf("c_cc[VDISCARD] = %i\n",c_cc[VDISCARD]);
	if(c_cc[VWERASE]) printf("c_cc[VWERASE] = %i\n",c_cc[VWERASE]);
	if(c_cc[VLNEXT]) printf("c_cc[VLNEXT] = %i\n",c_cc[VLNEXT]);
	if(c_cc[VEOL2]) printf("c_cc[VEOL2] = %i\n",c_cc[VEOL2]);
	printf("all other c_cc values 0\n");
}
/*----------------------------------------------------------
 translate_speed 
   accept:     speed (a member of the termios structure)
   perform:    print string given value
   return:     none.
   exceptions: none.
   comments:
----------------------------------------------------------*/ 
void translate_speed(int speed){
	switch(speed){
		case B0:  printf("0"); break;
		case B50:  printf("50"); break;
		case B75: printf("75"); break;
		case B110:  printf("110"); break;
		case B134: printf("134"); break;
		case B150:  printf("150"); break;
		case B200:  printf("200"); break;
		case B300:  printf("300"); break;
		case B600:  printf("600"); break;
		case B1200:  printf("1200"); break;
		case B1800:  printf("1800"); break;
		case B2400:  printf("2400"); break;
		case B4800:  printf("4800"); break;
		case B9600:  printf("9600"); break;
		case B19200:  printf("19200"); break;
		case B38400:  printf("38400"); break;
#ifdef B57600
		case B57600:  printf("57600"); break; 
#endif
#ifdef B115200
		case B115200:  printf("115200"); break; 
#endif
#ifdef B230400
		case B230400:  printf("230400"); break; 
#endif
#ifdef B460800
		case B460800:  printf("460800"); break; 
#endif
	}
}

/*----------------------------------------------------------
Termios_dumptermios
   accept:     
   perform:    transfer termios data from java side to native side
   return:     0 on success
               -1 on fail
   exceptions: none.
   comments:   This is starting to be limiting for users.  Need to handle
               possibility of multiple termios structs (see global ttyset 
               in/out) ?  do we call tcsetattr or go with the standard?
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_Termios_dumptermios(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
   )
{
/*
FIXME
	const char *str;
	jstring jstr;
*/
	jclass    Jtermios = GetObjectClass(Env, jobj);
	jfieldID c_iflag=GetStaticFieldID(Env, Jtermios, "c_iflag", "I");
	jfieldID c_oflag=GetStaticFieldID(Env, Jtermios, "c_oflag", "I");
	jfieldID c_cflag=GetStaticFieldID(Env, Jtermios, "c_cflag", "I");
	jfieldID c_lflag=GetStaticFieldID(Env, Jtermios, "c_lflag", "I");
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	jfieldID c_ispeed=GetStaticFieldID(Env, Jtermios, "c_ispeed", "I");
	jfieldID c_ospeed=GetStaticFieldID(Env, Jtermios, "c_ospeed", "I");
#else
#endif /* newer libc in linux */
	struct termios *ttyset=&cp[PortNumber].ttyset;

/*
FIXME
	jfieldID c_cc  =GetFieldID(Env, Jtermios, "c_cc", "Ljava/lang/String;");
#if defined(__linux__) && __GLIBC__ >= 2  
	jfieldID c_line=GetFieldID(Env, Jtermios, "c_line","Ljava/lang/String");
	jfieldID c_line=GetFieldID(Env, Jtermios, "c_line", "LC");
#endif
*/

	ttyset->c_iflag=GetStaticIntField(Env, Jtermios, c_iflag);
	ttyset->c_oflag=GetStaticIntField(Env, Jtermios, c_oflag);
	ttyset->c_cflag=GetStaticIntField(Env, Jtermios, c_cflag);
	ttyset->c_lflag=GetStaticIntField(Env, Jtermios, c_lflag);

/* close your eyes and this will look better 
FIXME
	jstr = GetObjectField(Env, jobj, c_cc);
	str = GetStringUTFChars(Env,jstr,0);
	sprintf(ttyset->c_cc,"%s",str);
	ReleaseStringUTFChars(Env, jstr, NULL);
#if defined(__linux__) && __GLIBC__ >= 2 
	ttyset->c_line=GetCharField(Env,jobj, c_line);
#endif
*/
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	cfsetispeed(ttyset,GetStaticIntField(Env, Jtermios, c_ispeed));
	cfsetospeed(ttyset,GetStaticIntField(Env, Jtermios, c_ospeed));
#endif /* newer libc in linux */
#ifdef __FreeBSD__
	if(cfsetspeed(&Port->ttyset, bps2speed_t(speed))<0) goto fail;
#endif /* freebsd */
	return(0);
}
/*----------------------------------------------------------
Termios_filltermios

   accept:     
   perform:    transfer termios data from native side to java side
   return:      0 on success
               -1 on fail
   exceptions: none.
   comments:   This is limiting for users.  Need to handle possibility of 
               multiple termios structs (see global ttyset in/out)
               FIXME there is a mess... see below

----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_Termios_filltermios(
        JNIEnv *Env,
        jobject jobj,
	jint PortNumber
   )
{
   /*jstring jstr; */
	jclass    Jtermios = GetObjectClass(Env, jobj);
	jfieldID  c_iflag = GetStaticFieldID(Env, Jtermios, "c_iflag", "I");
	jfieldID  c_oflag = GetStaticFieldID(Env, Jtermios, "c_oflag", "I");
	jfieldID  c_cflag = GetStaticFieldID(Env, Jtermios, "c_cflag", "I");
	jfieldID  c_lflag = GetStaticFieldID(Env, Jtermios, "c_lflag", "I");
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	jfieldID  c_ispeed = GetStaticFieldID(Env, Jtermios, "c_ispeed", "I");
	jfieldID  c_ospeed = GetStaticFieldID(Env, Jtermios, "c_ospeed", "I");
#else
#endif /* newer libc in linux */

	struct termios *ttyset=&cp[PortNumber].ttyset;
/* the pain.
FIXME
	jfieldID c_cc   =GetFieldID(Env, Jtermios, "c_cc", "Ljava/lang/String;");
#if defined(__linux__) && __GLIBC__ >= 2 
	jfieldID c_line =GetFieldID(Env,Jtermios,"c_line", "Ljava/lang/String");
#endif

	jstr = NewStringUTF(Env, ttyset->c_cc);
	SetObjectField(Env, jobj, c_cc, jstr);
	jstr = NewStringUTF(Env, ttyset->c_cc);
#if defined(__linux__) && __GLIBC__ >= 2  
	SetObjectField(Env,jobj, ttyset->c_line,jstr);
#endif
*/
	SetStaticIntField(Env, Jtermios, c_iflag, ttyset->c_iflag);
	SetStaticIntField(Env, Jtermios, c_oflag, ttyset->c_oflag);
	SetStaticIntField(Env, Jtermios, c_cflag, ttyset->c_cflag);
	SetStaticIntField(Env, Jtermios, c_lflag, ttyset->c_lflag);

   /* hmm  2 structures.  kernel handles it as one.  whats best to do?*/
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	SetStaticIntField(Env,Jtermios,c_ispeed, cfgetispeed(ttyset));
	SetStaticIntField(Env,Jtermios,c_ospeed, cfgetospeed(ttyset));
#else
#endif /* newer libc in linux */
	return(0);
}
/*----------------------------------------------------------
Termios_jstring2termios

   accept:      jstr (human readable termios define)
   perform:     convert java strings to termbits.h defines
   return:      -1 on error
                the #define for the string received on success.
   exceptions:  none.
   comments:    basically ran termbits.h through sed/paste
                this allows java to know about include files.
----------------------------------------------------------*/ 

JNIEXPORT jint JNICALL Java_gnu_io_Termios_jstring2termios(
        JNIEnv *Env,
        jobject jobj,
        jstring jstr
   )
{
	const char *javastring=GetStringUTFChars(Env,jstr,0);
	ReleaseStringUTFChars(Env, jstr, NULL);


#ifdef ALTWERASE
	else if(!strcmp("ALTWERASE", javastring )) return 	ALTWERASE;
#endif
#ifdef B0
	if(!strcmp("B0", javastring )) return 	B0;
#endif /* B0 */
#ifdef B110
	else if(!strcmp("B110", javastring )) return 	B110;
#endif /* B110 */
#ifdef B115200
	else if(!strcmp("B115200", javastring )) return 	B115200;
#endif /* B115200 */
#ifdef B1200
	else if(!strcmp("B1200", javastring )) return 	B1200;
#endif /* B1200 */
#ifdef B134
	else if(!strcmp("B134", javastring )) return 	B134;
#endif /* B134 */
#ifdef B150
	else if(!strcmp("B150", javastring )) return 	B150;
#endif /* B150 */
#ifdef B1800
	else if(!strcmp("B1800", javastring )) return 	B1800;
#endif /* B1800 */
#ifdef B19200
	else if(!strcmp("B19200", javastring )) return 	B19200;
#endif /* B19200 */
#ifdef B200
	else if(!strcmp("B200", javastring )) return 	B200;
#endif /* B200 */
#ifdef B2303400
	else if(!strcmp("B230400", javastring )) return 	B230400;
#endif /* B2303400 */
#ifdef BB2400
	else if(!strcmp("B2400", javastring )) return 	B2400;
#endif /* BB2400 */
#ifdef B300
	else if(!strcmp("B300", javastring )) return 	B300;
#endif /* B300 */
#ifdef B38400
	else if(!strcmp("B38400", javastring )) return 	B38400;
#endif /* B38400 */
#ifdef B460800
	else if(!strcmp("B460800", javastring )) return 	B460800;
#endif /* B460800 */
#ifdef B4800
	else if(!strcmp("B4800", javastring )) return 	B4800;
#endif /* B4800 */
#ifdef B50
	else if(!strcmp("B50", javastring )) return 	B50;
#endif /* B50 */
#ifdef B57600
	else if(!strcmp("B57600 ", javastring )) return 	B57600 ;
#endif /* B57600 */
#ifdef B600
	else if(!strcmp("B600", javastring )) return 	B600;
#endif /* B600 */
#ifdef B75
	else if(!strcmp("B75", javastring )) return 	B75;
#endif /* B75 */
#ifdef B9600
	else if(!strcmp("B9600", javastring )) return 	B9600;
#endif /* B9600 */
#ifdef BRKINT
	else if(!strcmp("BRKINT", javastring )) return 	BRKINT;
#endif /* BRKINT */
#ifdef BS0
	else if(!strcmp("BS0", javastring )) return 	BS0;
#endif /* BS0 */
#ifdef BS1
	else if(!strcmp("BS1", javastring )) return 	BS1;
#endif /* BS1 */
#ifdef BSDLY
	else if(!strcmp("BSDLY", javastring )) return 	BSDLY;
#endif /* BSDLY */
#ifdef CBAUD
	else if(!strcmp("CBAUD", javastring )) return 	CBAUD;
#endif /* CBAUD */
#ifdef CBAUDEX
	else if(!strcmp("CBAUDEX", javastring )) return 	CBAUDEX;
#endif /* CBAUDEX appears to be  unused on linux. */
#ifdef CCTS_IFLOW
	else if(!strcmp("CCTS_IFLOW", javastring )) return 	CCTS_IFLOW;
#endif
#ifdef CCTS_OFLOW
	else if(!strcmp("CCTS_OFLOW", javastring )) return 	CCTS_OFLOW;
#endif
#ifdef CIBAUD
	else if(!strcmp("CIBAUD", javastring )) return 	CIBAUD; 
#endif /* CIBAUD */
#ifdef CLOCAL
	else if(!strcmp("CLOCAL", javastring )) return 	CLOCAL;
#endif /* CLOCAL */
#ifdef CR0
	else if(!strcmp("CR0", javastring )) return 	CR0;
#endif /* CR0 */
#ifdef CR1
	else if(!strcmp("CR1", javastring )) return 	CR1;
#endif /* CR1 */
#ifdef CR2
	else if(!strcmp("CR2", javastring )) return 	CR2;
#endif /* CR2 */
#ifdef CR3
	else if(!strcmp("CR3", javastring )) return 	CR3;
#endif /* CR3 */
#ifdef CRDLY
   else if(!strcmp("CRDLY", javastring )) return 	CRDLY;
#endif /* CRDLY */
#ifdef CREAD
	else if(!strcmp("CREAD", javastring )) return 	CREAD;
#endif /* CREAD */
#ifdef CRTSCTS
	else if(!strcmp("CRTSCTS", javastring )) return 	CRTSCTS;
#endif /* CRTSCTS */
#ifdef CS5
	else if(!strcmp("CS5", javastring )) return 	CS5;
#endif /* CS5 */
#ifdef CS6
	else if(!strcmp("CS6", javastring )) return 	CS6;
#endif /* CS6 */
#ifdef CS7
	else if(!strcmp("CS7", javastring )) return 	CS7;
#endif /* CS7 */
#ifdef CS8
	else if(!strcmp("CS8", javastring )) return 	CS8;
#endif /* CS8 */
#ifdef CSIZE
	else if(!strcmp("CSIZE", javastring )) return 	CSIZE;
#endif /* CSIZE */
#ifdef CSTOPB
	else if(!strcmp("CSTOPB", javastring )) return 	CSTOPB;
#endif /* CSTOPB */
#ifdef ECHO
	else if(!strcmp("ECHO", javastring )) return 	ECHO;
#endif /* ECHO */
#ifdef ECHOCTL
	else if(!strcmp("ECHOCTL", javastring )) return 	ECHOCTL;
#endif /* ECHOCTL */
#ifdef ECHOE
	else if(!strcmp("ECHOE", javastring )) return 	ECHOE;
#endif /* ECHOE */
#ifdef ECHOK
	else if(!strcmp("ECHOK", javastring )) return 	ECHOK;
#endif /* ECHOK */
#ifdef ECHOKE
	else if(!strcmp("ECHOKE", javastring )) return 	ECHOKE;
#endif /* ECHOKE */
#ifdef ECHONL
	else if(!strcmp("ECHONL", javastring )) return 	ECHONL;
#endif /* ECHONL */
#ifdef ECHOPRT
	else if(!strcmp("ECHOPRT", javastring )) return 	ECHOPRT;
#endif /* ECHOPRT */
#ifdef EXTA
	else if(!strcmp("EXTA", javastring )) return 	EXTA;
#endif /* EXTA */
#ifdef EXTB
	else if(!strcmp("EXTB", javastring )) return 	EXTB;
#endif /* EXTB */
#ifdef EXTPROC
	else if(!strcmp("EXTPROC", javastring )) return 	EXTPROC;
#endif
#ifdef FF0
	else if(!strcmp("FF0", javastring )) return 	FF0;
#endif /* FF0 */
#ifdef FF1
	else if(!strcmp("FF1", javastring )) return 	FF1;
#endif /* FF1 */
#ifdef FFDLY
	else if(!strcmp("FFDLY", javastring )) return 	FFDLY;
#endif /* FFDLY */
#ifdef FLUSHO
	else if(!strcmp("FLUSHO", javastring )) return 	FLUSHO;
#endif /* FLUSHO */
#ifdef HUPCL
	else if(!strcmp("HUPCL", javastring )) return 	HUPCL;
#endif /* HUPCL */
#ifdef ICANON
	else if(!strcmp("ICANON", javastring )) return 	ICANON;
#endif /* ICANON */
#ifdef ICRNL
	else if(!strcmp("ICRNL", javastring )) return 	ICRNL;
#endif /* ICRNL */
#ifdef IEXTEN
	else if(!strcmp("IEXTEN", javastring )) return 	IEXTEN;
#endif /* IEXTEN */
#ifdef IGNBRK
	else if(!strcmp("IGNBRK", javastring )) return 	IGNBRK;
#endif /* IGNBRK */
#ifdef IGNCR
	else if(!strcmp("IGNCR", javastring )) return 	IGNCR;
#endif /* IGNCR */
#ifdef IGNPAR
	else if(!strcmp("IGNPAR", javastring )) return 	IGNPAR;
#endif /* IGNPAR */
#ifdef IMAXBEL
	else if(!strcmp("IMAXBEL", javastring )) return 	IMAXBEL;
#endif /* IMAXBEL */
#ifdef INLCR
	else if(!strcmp("INLCR", javastring )) return 	INLCR;
#endif /* INLCR */
#ifdef INPCK
	else if(!strcmp("INPCK", javastring )) return 	INPCK;
#endif /* INPCK */
#ifdef ISIG
	else if(!strcmp("ISIG", javastring )) return 	ISIG;
#endif /* ISIG */
#ifdef ISTRIP
	else if(!strcmp("ISTRIP", javastring )) return 	ISTRIP;
#endif /* ISTRIP */
#ifdef IUCLC /* legacy support */
	else if(!strcmp("IUCLC", javastring )) return 	IUCLC;
#endif /* IUCLC */
#ifdef IXANY
	else if(!strcmp("IXANY", javastring )) return 	IXANY;
#endif /* IXANY */
#ifdef IXOFF
	else if(!strcmp("IXOFF", javastring )) return 	IXOFF;
#endif /* IXOFF */
#ifdef IXON
	else if(!strcmp("IXON", javastring )) return 	IXON;
#endif /* IXON */
#ifdef MDMBUF
	else if(!strcmp("MDMBUF", javastring )) return 	MDMBUF;
#endif
#ifdef NL0
	else if(!strcmp("NL0", javastring )) return 	NL0;
#endif /* NL0 */
#ifdef NL1
	else if(!strcmp("NL1", javastring )) return 	NL1;
#endif /* NL1 */
#ifdef NLDLY
	else if(!strcmp("NLDLY", javastring )) return 	NLDLY;
#endif /* NLDLY */
#ifdef NOFLSH
	else if(!strcmp("NOFLSH", javastring )) return 	NOFLSH;
#endif /* NOFLSH */
#ifdef NOKERNINFO
	else if(!strcmp("NOKERNINFO", javastring )) return 	NOKERNINFO;
#endif
#ifdef OCRNL
	else if(!strcmp("OCRNL", javastring )) return 	OCRNL;
#endif /* OCRNL */
#ifdef OFDEL
	else if(!strcmp("OFDEL", javastring )) return 	OFDEL;
#endif /* OFDEL */
#ifdef OFILL
	else if(!strcmp("OFILL", javastring )) return 	OFILL;
#endif /* OFILL */
#ifdef OLCUC /* legacy support */
	else if(!strcmp("OLCUC", javastring )) return 	OLCUC;
#endif /* OLCUC */
#ifdef ONLCR
	else if(!strcmp("ONLCR", javastring )) return 	ONLCR;
#endif /* ONLCR */
#ifdef ONLRET
	else if(!strcmp("ONLRET", javastring )) return 	ONLRET;
#endif /* ONLRET */
#ifdef ONOCR
	else if(!strcmp("ONOCR", javastring )) return 	ONOCR;
#endif /* ONOCR */
#ifdef ONOEOT
	else if(!strcmp("ONOEOT", javastring )) return 	ONOEOT;
#endif
#ifdef OPOST
	else if(!strcmp("OPOST", javastring )) return 	OPOST;
#endif /* OPOST */
#ifdef OXTABS
	else if(!strcmp("OXTABS", javastring )) return 	OXTABS;
#endif
#ifdef PARENB
	else if(!strcmp("PARENB", javastring )) return 	PARENB;
#endif /* PARENB */
#ifdef PARMRK
	else if(!strcmp("PARMRK", javastring )) return 	PARMRK;
#endif /* PARMRK */
#ifdef PARODD
	else if(!strcmp("PARODD", javastring )) return 	PARODD;
#endif /* PARODD */
#ifdef PENDIN
	else if(!strcmp("PENDIN", javastring )) return 	PENDIN;
#endif /* PENDIN */
#ifdef TAB0
	else if(!strcmp("TAB0", javastring )) return 	TAB0;
#endif /* TAB0 */
#ifdef TAB1
	else if(!strcmp("TAB1", javastring )) return 	TAB1;
#endif /* TAB1 */
#ifdef TAB2
	else if(!strcmp("TAB2", javastring )) return 	TAB2;
#endif /* TAB2 */
#ifdef TAB3
	else if(!strcmp("TAB3", javastring )) return 	TAB3;
#endif /* TAB3 */
#ifdef TABDLY
	else if(!strcmp("TABDLY", javastring )) return 	TABDLY;
#endif /* TABDLY */
#ifdef TCIFLUSH
	else if(!strcmp("TCIFLUSH", javastring )) return 	TCIFLUSH;
#endif /* TCIFLUSH */
#ifdef TCIOFF
	else if(!strcmp("TCIOFF", javastring )) return 	TCIOFF;
#endif /* TCIOFF */
#ifdef TCIOFLUSH
	else if(!strcmp("TCIOFLUSH", javastring )) return 	TCIOFLUSH;
#endif /* TCIOFLUSH */
#ifdef TCION
	else if(!strcmp("TCION", javastring )) return 	TCION;
#endif /* TCION */
#ifdef TCIOFLUSH
	else if(!strcmp("TCOFLUSH", javastring )) return 	TCOFLUSH;
#endif /* TCIOFLUSH */
#ifdef TCOOFF
	else if(!strcmp("TCOOFF", javastring )) return 	TCOOFF;
#endif /* TCOOFF */
#ifdef TCOON
	else if(!strcmp("TCOON", javastring )) return 	TCOON;
#endif /* TCOON */
#ifdef TCSADRAIN
	else if(!strcmp("TCSADRAIN", javastring )) return 	TCSADRAIN;
#endif /* TCSADRAIN */
#ifdef TCSAFLUSH
	else if(!strcmp("TCSAFLUSH", javastring )) return 	TCSAFLUSH;
#endif /* TCSAFLUSH */
#ifdef TCSANOW
	else if(!strcmp("TCSANOW", javastring )) return 	TCSANOW;
#endif /* TCSANOW */
#ifdef TOSTOP
	else if(!strcmp("TOSTOP", javastring )) return 	TOSTOP;
#endif /* TOSTOP */
#ifdef VDISCARD
	else if(!strcmp("VDISCARD", javastring )) return 	VDISCARD;
#endif /* VDISCARD */
#ifdef VEOF
	else if(!strcmp("VEOF", javastring )) return 	VEOF;
#endif /* VEOF */
#ifdef VEOL
	else if(!strcmp("VEOL", javastring )) return 	VEOL;
#endif /* VEOL */
#ifdef VEOL2
	else if(!strcmp("VEOL2", javastring )) return 	VEOL2;
#endif /* VEOL2 */
#ifdef VERASE
	else if(!strcmp("VERASE", javastring )) return 	VERASE;
#endif /* VERASE */
#ifdef VINTR
	else if(!strcmp("VINTR", javastring )) return 	VINTR;
#endif /* VINTR */
#ifdef VKILL
	else if(!strcmp("VKILL", javastring )) return 	VKILL;
#endif /* VKILL */
#ifdef VLNEXT
	else if(!strcmp("VLNEXT", javastring )) return 	VLNEXT;
#endif /* VLNEXT */
#ifdef VMIN
	else if(!strcmp("VMIN", javastring )) return 	VMIN;
#endif /* VMIN */
#ifdef VQUIT
	else if(!strcmp("VQUIT", javastring )) return 	VQUIT;
#endif /* VQUIT */
#ifdef VREPRINT
	else if(!strcmp("VREPRINT", javastring )) return 	VREPRINT;
#endif /* VREPRINT */
#ifdef VSTART
	else if(!strcmp("VSTART", javastring )) return 	VSTART;
#endif /* VSTART */
#ifdef VSTOP
	else if(!strcmp("VSTOP", javastring )) return 	VSTOP;
#endif /* VSTOP */
#ifdef VSUSP
	else if(!strcmp("VSUSP", javastring )) return 	VSUSP;
#endif /* VSUSP */
#ifdef VSWTC
	else if(!strcmp("VSWTC", javastring )) return 	VSWTC;
#endif /* VSWTC */
#ifdef VT0
	else if(!strcmp("VT0", javastring )) return 	VT0;
#endif /* VT0 */
#ifdef VT1
	else if(!strcmp("VT1", javastring )) return 	VT1;
#endif /* VT1 */
#ifdef VTDLY
	else if(!strcmp("VTDLY", javastring )) return 	VTDLY;
#endif /* VTDLY */
#ifdef VTIME
	else if(!strcmp("VTIME", javastring )) return 	VTIME;
#endif /* VTIME */
#ifdef VWERASE
	else if(!strcmp("VWERASE", javastring )) return 	VWERASE;
#endif /* VWERASE */
#ifdef XCASE /* legacy support */
	else if(!strcmp("XCASE", javastring )) return 	XCASE;
#endif /* XCASE */
#ifdef XTABS
	else if(!strcmp("XTABS", javastring )) return 	XTABS;
#endif /* XTABS */
	else return(-1);
}
/*----------------------------------------------------------
Termios_haveiospeed

   accept:      none.
   perform:     find out if the termios structure has iospeed.
   return:      0 if c_i/ospeed is not available
                1 if c_i/ospeed is available
   exceptions:  none.
   comments:    The problem we are trying to deal with is libc5 on linux does 
                not have c_ispeed/c_ospeed in the termios structure
                if other OS's have this problem deal with it here.
----------------------------------------------------------*/ 

JNIEXPORT jboolean JNICALL Java_gnu_io_Termios_haveiospeed(
        JNIEnv *Env,
        jobject jobj
   )
{
#if defined(__linux__) && __GLIBC__ >= 2  /* newer libc in linux */
	return(1);
#else
	return(0);
#endif /* newer libc in linux */
}
/* printf("Serial port initialized %d, %d \n",cfgetospeed(&cp[PortNumber].ttyset), cfgetispeed(&cp[PortNumber].tyset));
*/

