/*-------------------------------------------------------------------------
|   rxtx is a native interface to serial ports in java.
|   Copyright 1997, 1998, 1999 by Trent Jarvi jarvi@ezlink.com.
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------
   win32 support has been removed.  There was not enough developer interest
   if your interested in working on win32 support contact jarvi@ezlink.com
--------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------

The Roadmap.

Find a problem? toss a FIXME here and explain in comments for the function.
Fixed something? Remove the comment and the note here if the function is fine.

### Common to SerialOutputStream* and SerialInputStream*

[hi level speed control]
SetSpeed                   set the termios structure speed to speed. 
GetSpeed                   get the speed from the termios structure. 

   For the following SerialInputStream functions just call SerialOutputStream.  

open                       open the device.                           
close                      close the device.
tcsetattr                  Get the termios structure to the kernel.  
tcsendbreak                send 0 bytes for actual time or >= 0.25 secFIXME?
tcdrain                    return after all pending data is written.
tcflush                    flush input/output depending on selector. 
tcflow                     try to start/stop input/output.
cfmakeraw                  set the termios structure into a sane mode.
tcgetpgrp                  gets process group ID.                    
tcsetpgrp                  sets process group ID.                     
isCD                       is CD set?  Carrier Detect.
isCTS                      Is CTS set? Clear To Send.
isDSR                      is DSR set? Data Set Ready.
isRI                       is RI set?  Ring Indicator.
isRTS                      is RTS set?  Request To Send.
isDTR                      is DTR set? Data Terminal Ready.
setRTS                     set RTS  Set Request To Send.
InitPortArray              sets the port struct array to sane values first time.
                                     
### Unique to SerialOutputStream*

openAppend                 ioexception.                               
nativewrite                write a byte to the device.
writeBytes                 write offset zeros.  write length bytes    

[low level speed control]
cfgetospeed                termios functions
cfsetospeed                set the termios structure o_speed value.

setoutputbuffersize        Set the output buffer size.
getoutputbuffersize        return the output buffer size.

### Unique to SerialInputStream*

nativeread                 read bytes from a file descriptor          
readBytes                  skip offset bytes and read Length bytes FIXME
skip                       toss out jl bytes from the input stream.   
[low level speed control]
cfgetispeed                Checks the termios structure for the input speed.
cfsetispeed                set the termios structure output speed value.

nativeavailable            find out if bytes are available for reading.
noblock                    hardware flow control
setHWFC                    set hardware flow control to state.
setDTR                     set DTR. Data Terminal Ready.
setinputbuffersize         set the input buffer size
getinputbuffersize         return the input buffer size
enablereceiveframing       enable receive framing/set framing byte
disablereceiveframing      disable receive framing/set framing byte=0
getreceiveframingbyte      return the framing byte
isreceiveframingenabled    if framingbyte !=0 return true otherwise return false
gettimeout                 return the timeout value in milliseconds
settimeout                 set the timeout value in milliseconds
enablereceivethreshold     set the enable receive threshold
getreceivethreshold        return the receive threshold
disablereceivethreshold    set receive threshold to 0.
isreceivethresholdenabled  set receive threshold to 0.

### Unique to Termios*

strlenccc                  get the string length of termios.c_cc      
ShowNativeTermios          print termios structure from native side   
dumptermios                transfer termios from java to native       FIXME
filltermios                transfer termios data from native to java  FIXME
jstring2termios            convert java strings to termbits.h defines
haveiospeed                find out if the termios structure has iospeed.

### Misc functions.

show_ctl                   hardware flow control.                     
makeraw                    sets port in sane setting.
bps2speed_t                Convert bits-per-second to a speed_t constant.
speed_t2bps		   Convert a speed_t constant to a bits-per-second.
IOException                Thows an io exception to java.
translate_ciflag           print termios.c_iflag
translate_coflag           print termios.c_oflag
translate_ccflag           print termios.c_cflag
translate_clflag           print termios.c_lflag
translate_c_cc             print the set flags in c_cc control characters
translate_speed            print termios.speed
timeval_subtract           

End The Roadmap.
--------------------------------------------------------------------------*/
/*
  Disable debugging if NDEBUG is defined, following the conventions 
  set by assert(). 
*/


/* values for BLOCKING flag */
#define BLOCKING 0
#define NONBLOCK 1

/* value for event notification */
#define FRAME_ERROR 8; 
#define BREAK_ERROR 4; 
#define PARITY_ERROR 2;
#define OVERRUN_ERROR 1;

/*
Flow Control defines inspired by reading how mgetty by Gert Doering does it
*/

#ifdef CRTSCTS
#define HARDWARE_FLOW_CONTROL CRTSCTS
#else
#       ifdef CCTS_OFLOW
#       define HARDWARE_FLOW_CONTROL CCTS_OFLOW|CRST_IFLOW
#       else
#               ifdef RTSFLOW
#               define HARDWARE_FLOW_CONTROL RTSFLOW|CTSFLOW
#               else
#                       ifdef CRTSFL
#                       define HARDWARE_FLOW_CONTROL CRTSFL
#                       else
#                               ifdef CTSCD
#                               define HARDWARE_FLOW_CONTROL CTSCD
#                               else
#                                       define HARDWARE_FLOW_CONTROL 0
#                               endif
#                       endif
#               endif
#       endif
#endif


/* PROTOTYPES */
int speed_t2bps(int);
int bps2speed_t(int);
void IOException( JNIEnv *Env,char *function, char *msg); 
void makeraw(struct termios *ttyset);
void translate_ciflag(int c_iflag);
void translate_coflag(int c_oflag);
void translate_ccflag(int c_cflag);
void translate_clflag(int c_lflag);
void translate_c_cc(char *c_cc);
void translate_speed(int speed);
int timeval_subtract(struct timeval *result,struct timeval *x,struct timeval *y);
void toss_error(JNIEnv *Env,char *msg1, char *msg2);
void toss_debug_blocking(char *function_name, int PortNumber);
void toss_debug(char *function_name, int PortNumber);





/*----------------------------------------------------------

This is for multiple ports

----------------------------------------------------------*/ 

static struct commport {
	int available;        /* used to indicate if this struct is used */
	int fd;               /* file descriptor returned by open */
	int fd_is_needed;    /* used to negotiate use of fd between threads */
	int number_of_times_opened;/* number of times port is opened */
	char name[80];        /* name passed to open ie.. /dev/ttyS1 */
	char no_block;        /* flag to indicate the the user wants blocking */
	int hardwarefc;       /* flag to indicate hardware flowcontrol */
/*	int CD_state;          last state of CD for hardware flowcontrol */
#ifdef WAYNE
	int JTIMEOUT;         /* set by enableReceiveTimeout() in java */ 
	int THRESHOLD;        /* set bye enableReceiveThreshold() in java */
	int BAUDRATE;         /* value passed to SetSpeed, default is 9600  */
#endif
	int last_TIOCMGET_result; /* used by show_ctl() */
	int OutputBufferSize; /* not implemented in write */
	int InputBufferSize;  /* not inmplemented in read */
	char FrameByte;       /* not inmplemented in read */
#ifdef EVENTNOTIFICATION
#if defined(__linux__)
	struct serial_icounter_struct sis;
	struct serial_icounter_struct osis;
	struct serial_struct ss;
	struct serial_struct oss;
#endif
#endif
	struct termios ttyset;/* termios info */
	struct termios ttyorig;/* termios info */

} cp[MAXPORTS];

#ifdef GREENTHREAD_WORKAROUND  /* use native threads instead of this */
#define open __open            /* probably glibc specific */
#define read __read
#endif

/*
  Disable debugging if NDEBUG is defined, following the conventions
  set by assert(). 
*/
#ifdef NDEBUG
#undef DEBUG
#endif
#define GetObjectField (*Env)->GetObjectField
#define GetStaticFieldID (*Env)->GetStaticFieldID
#define GetObjectClass (*Env)->GetObjectClass
#define GetStringUTFChars (*Env)->GetStringUTFChars
#define GetByteArrayElements (*Env)->GetByteArrayElements
#define ReleaseStringUTFChars (*Env)->ReleaseStringUTFChars
#define ReleaseByteArrayElements (*Env)->ReleaseByteArrayElements
#define SetStaticIntField (*Env)->SetStaticIntField
#define FindClass (*Env)->FindClass
#define ThrowNew (*Env)->ThrowNew
#define GetFieldID (*Env)->GetFieldID
#define GetStaticIntField (*Env)->GetStaticIntField
#define GetCharField (*Env)->GetCharField
#define NewStringUTF (*Env)->NewStringUTF
#define SetObjectField (*Env)->SetObjectField


