/*-------------------------------------------------------------------------
|   rxtx is a native interface to serial ports in java.
|   Copyright 1997, 1998, 1999 by Trent Jarvi jarvi@ezlink.com.
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
package gnu.io;

import java.io.InputStream;
import java.io.IOException;
import java.io.FileDescriptor;
import java.io.FileNotFoundException;
import java.io.File;

public class SerialInputStream extends InputStream 
{
	private final static boolean debug = false;

	private FD fd;        
	static int FileD=-1; // Used by the native methods. needed?
	int PortNumber=-1; // Used by the native methods.

	public SerialInputStream(String filename) throws FileNotFoundException 
	{
		if( debug ) System.out.println( "SerialInputStream( \"" + filename + "\" )" );
		InitPortArray();

		SecurityManager Check = System.getSecurityManager();
		if (Check != null) {
			Check.checkWrite(filename);
        	} else {
			if( debug ) System.out.println("Security manager not used");
		}

		try {
			fd = new FD();
			/* FIXME 
			   presumably some magic happens between... 
			   need to call setFD() from native side 
			   then use the FD in future calls (not critical yet) */
			PortNumber=open(filename);
		} catch (IOException e) {
			throw new FileNotFoundException(filename);
		}
		if(PortNumber<0)
			throw new FileNotFoundException(filename);
	}
	/* This one makes no sense to me  for serial input/output*/
	public SerialInputStream(String filename, boolean openappend) throws IOException 
	{
		System.out.println("SerialInputStream(String filename, boolean)");
		System.out.println("openappend) not implemented");
		System.out.println("Look at the source of ");
		System.out.println("input streams for hints");
		throw new FileNotFoundException(filename);
	}
	public SerialInputStream(File file) throws IOException {
		this(file.getPath());
	}
	/*
	I dont think we are ready for this.  Have to build a fd class so we can
	access the private fd.   Just use the filename... per above.
	*/
	public SerialInputStream(FileDescriptor Inputfd) 
	{
		System.out.println("SerialInputStream(FileDescriptor Inputfd)");
		System.out.println("not implemented.  Look at the source of ");
		System.out.println("input streams for hints");
		throw new NullPointerException();
	}
	public native void SetSpeed(int PortNumber, int speed) throws IOException;
	public native int GetSpeed(int PortNumber) throws IOException;
	public native int tcsetattr (int PortNumber, int actions);
	public native int tcflow (int PortNumber, int action);
	public native void cfmakeraw (int PortNumber);
	public native int cfgetispeed (int PortNumber);
	public native int cfsetispeed (int PortNumber, int speed);
	public native int tcgetpgrp (int PortNumber);
	public native int tcsetpgrp (int PortNumber, int pgrpid);
	public native byte catchsigio();
	private native void InitPortArray();
	private native int open(String name) throws IOException;
	public native int nativeread(int PortNumber) throws IOException;
	private native int readBytes(int PortNumber, byte b[], int off, int len) throws IOException;
	public native long skip(int PortNumber, long n) throws IOException;
	public native int nativeavailable(int PortNumber ) throws IOException;
	public native void noblock(int PortNumber, int state); 
	public native void setHWFC(int PortNumber, int state); 
        public native boolean isCTS(int PortNumber);
	public native boolean isDSR(int PortNumber );
        public native boolean isCD(int PortNumber);
	public native boolean isRI(int PortNumber );
	public native boolean isRTS(int PortNumber );
	public native boolean isFrame(int PortNumber );
	public native boolean isEmpty(int PortNumber );
	public native boolean isOverrun(int PortNumber );
	public native boolean isBreak(int PortNumber );
	public native boolean isParity(int PortNumber );
	public native void setDTR(int PortNumber, boolean state);
        public native void setinputbuffersize(int PortNumber, int size);
        public native int getinputbuffersize(int PortNumber);
        public native void enablereceiveframing(int PortNumber, int Byte);
        public native void disablereceiveframing(int PortNumber);
        public native int getreceiveframingbyte(int PortNumber);
        public native boolean isreceiveframingenabled(int PortNumber);
        public native int gettimeout(int PortNumber );
        public native void settimeout(int PortNumber, int time);
        public native void enablereceivethreshold(int PortNumber, int time);
        public native void disablereceivethreshold(int PortNumber );
        public native int getreceivethreshold(int PortNumber );
        public native boolean isreceivethresholdenabled(int PortNumber);
	public native void close(int PortNumber) throws IOException;

	public void close() throws IOException
	{
		close(PortNumber);
	}
	public int read() throws IOException{
		if( debug ) System.out.println( "Java->read()" );
		return nativeread(PortNumber);
	}
	public int read(byte b[]) throws IOException {
		if( debug ) System.out.println( "Java->read(b[])" );
		return readBytes(PortNumber, b, 0, b.length);
	}
	public int read(byte b[], int off, int len) throws IOException {
		if( debug ) System.out.println( "Java->read(b[],off)" );
		return readBytes(PortNumber, b, off, len);
	}
	public int available() throws IOException {
		return (nativeavailable(PortNumber ));
	}
	public final FD getFD() throws IOException 
	{
		if (fd != null) return fd;
			throw new IOException();
	}
	public final void setFD(FD fd) throws IOException
	{
		if (fd != null) this.fd = fd;
			throw new IOException();
	}
	protected void finalize() throws IOException 
	{
		if (fd != null) 
			close(PortNumber);
	}
	static {
		System.loadLibrary("Serial");
	}
}
