/*-------------------------------------------------------------------------
|   rxtx is a native interface to serial ports in java.
|   Copyright 1997, 1998, 1999 by Trent Jarvi jarvi@ezlink.com.
|
|   This library is free software; you can redistribute it and/or
|   modify it under the terms of the GNU Library General Public
|   License as published by the Free Software Foundation; either
|   version 2 of the License, or (at your option) any later version.
|
|   This library is distributed in the hope that it will be useful,
|   but WITHOUT ANY WARRANTY; without even the implied warranty of
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
|   Library General Public License for more details.
|
|   You should have received a copy of the GNU Library General Public
|   License along with this library; if not, write to the Free
|   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------*/
package gnu.io;

import java.io.OutputStream;
import java.io.IOException;
import java.io.FileDescriptor;
import java.io.FileNotFoundException;
import java.io.File;

public class SerialOutputStream extends OutputStream
{
	private final static boolean debug = false;
	private FD fd;
	static int FileD=-1; // Used by native methods.
	int PortNumber=-1; // Used by native methods.

	public SerialOutputStream(String filename) throws FileNotFoundException
	{
		if( debug ) System.out.println( "SerialOutputStream( \"" + filename + "\" )" );
		InitPortArray();

		SecurityManager Check = System.getSecurityManager();
		if (Check != null) {
			Check.checkWrite(filename);
		} else {
		// Ignored unless debugging on because regular users doen't care
		// if a security manager is present.
			if( debug ) {
				System.out.println("Security manager not used");
			}
		}

		try {
			fd = new FD();
			/* FIXME */
			/* presumably some magic happens between... */
			/* need to call setFD() from native side */
			/* then use the FD in future calls (not critical yet) */

			PortNumber=open(filename);
		} catch (IOException e) {
			throw new FileNotFoundException(filename);
		}
		if(PortNumber<0)
			throw new FileNotFoundException(filename);

	}
	/* This one makes no sense to me  for serial input/output*/
	public SerialOutputStream(String filename, boolean openappend) throws IOException 
	{
		System.out.println("SerialOutputStream(String filename, boolean)");
		System.out.println("openappend) not implemented");
		System.out.println("Look at the source of ");
		System.out.println("input streams for hints");
		throw new FileNotFoundException(filename);
	}
	public SerialOutputStream(File file) throws IOException {
		this(file.getPath());
	}
	/*
	I dont think we are ready for this.  Have to build a fd class so we can
	access the private fd.   Just use the filename... per above.
	*/
	public SerialOutputStream(FileDescriptor Inputfd) {
		System.out.println("SerialInputStream(FileDescriptor Inputfd)");
		System.out.println("not implemented");
		System.out.println("Look at the source of ");
		System.out.println("input streams for hints");
		throw new NullPointerException();
	}
	/* unique to Serial */
	public native void SetSpeed(int PortNumber, int speed) throws IOException;
	public native int GetSpeed(int PortNumber) throws IOException;

	/*  TERMIOS(3) Support  */

	public native int tcsetattr (int PortNumber,int actions);
	public native int tcsendbreak (int PortNumber,int duration);
	public native int tcdrain (int PortNumber);
	public native int tcflush (int PortNumber,int selector);
	public native int tcflow (int PortNumber,int action);
	public native void cfmakeraw (int PortNumber);
	public native int cfgetospeed (int PortNumber);
	public native int cfsetospeed (int PortNumber,int speed);
	public native int tcgetpgrp (int PortNumber);
	public native int tcsetpgrp (int PortNumber,int pgrpid);
	
	// special fns to handle modem line queries

	public native boolean isDTR(int PortNumber);
	public native void setRTS(int PortNumber,boolean state);
	
        public native void setoutputbuffersize(int PortNumber,int size);
        public native int getoutputbuffersize(int PortNumber);
 	public native void InitPortArray();
	/* normal stream stuff */

	private native int open(String name) throws IOException;
	private native void openAppend(String name) throws IOException;
	public void write(int b) throws IOException 
	{
		if( debug ) System.out.println( "Java->write(b)" );
		nativewrite(PortNumber, b);
	}
	public native void nativewrite(int PortNumber, int b) throws IOException;
	private native void writeBytes(int PortNumber, byte b[], int off, int len) throws IOException;
	public void write(byte b[]) throws IOException 
	{
		if( debug ) System.out.println( "Java->write(b[])" );
		writeBytes(PortNumber, b, 0, b.length);
	}
	public void write(byte b[], int off, int len) throws IOException 
	{
		if( debug ) System.out.println( "Java->write(b[],off)" );
		writeBytes(PortNumber, b, off, len);
	}

	public void close() throws IOException
	{
		close(PortNumber);
	}
	public native void close(int PortNumber) throws IOException;
	public final FD getFD()  throws IOException 
	{
		if (fd != null) return fd;
		throw new IOException();
	}
	public final void setFD(FD fd) throws IOException
	{
		if (fd != null) this.fd = fd;
		throw new IOException();
	}
	protected void finalize() throws IOException 
	{
		if (fd != null) {
			flush();
			close(PortNumber);
		}
	}
	static {
		System.loadLibrary("Serial");
	}
}
